import pandas as pd
from typing import Dict, List
import logging
from continual.rpc.management.v1 import management_types_pb2, types
from continual.python.sdk.query import Query
from continual.python.sdk.features.featurestore import get_featurestore
from continual.services.compute_services.constants import POSTGRES_DB_URL


class Dataset:
    def __init__(
        self,
        model: management_types_pb2.Model,
        featureset_map: Dict[str, management_types_pb2.FeatureSet],
        data_store: management_types_pb2.DataStore,
    ):
        self._input_features = []
        self._target = None
        self._query: Query = None
        self._data_store = data_store

        self._get_feature_metadata(model=model, featureset_map=featureset_map)

    @property
    def input_features(self) -> List[types.ColumnConfig]:
        return self._input_features

    @property
    def target(self) -> types.ColumnConfig:
        return self._target

    @property
    def query(self) -> Query:
        return self._query

    @property
    def time_index(self) -> str:
        return self._query.get_time_index_name()

    @property
    def index(self) -> str:
        return self._query.get_index_name()

    @property
    def split_field(self) -> str:
        return self._query.get_split_name()

    @property
    def inferred_problem_type(self) -> str:
        """
        Gets problem type based off of outputs only
        """
        if self._target.type == types.FieldType.CATEGORICAL:
            return "multiclass_classification"
        if self._target.type == types.FieldType.BOOLEAN:
            return "binary_classification"
        if self._target.type == types.FieldType.NUMBER:
            return "regression"
        if self._target.type == types.FieldType.TEXT:
            return "text_prediction"

        raise ValueError(f"Unknown problem type for {self._target}")

    def get_data(
        self,
        local: bool = False,
        allrows: bool = False,
        database_url: str = POSTGRES_DB_URL,
    ) -> pd.DataFrame:
        """
        Returns data as pandas dataframe for this dataset
        """

        fs = get_featurestore(
            self._data_store, cached=False, cfg={"DatabaseURL": database_url}
        )
        qiter = fs.fetch_training_data(
            self._query.to_dict(),
            all_instances=allrows,
            local=local,
        )
        if local:
            return qiter

        all_dfs = []

        try:
            df = next(qiter)
            df.columns = df.columns.str.lower()
            cols = [c for c in df.columns if c.endswith("timerank")]
            if "finalrank" in df.columns:
                cols.append("finalrank")

            while df is not None:
                df.columns = df.columns.str.lower()
                if len(cols) > 0:
                    df.drop(cols, axis=1, inplace=True)
                all_dfs.append(df)
                df = next(qiter)
        except StopIteration:
            pass

        if fs:
            fs.close()

        if len(all_dfs) > 0:
            return pd.concat(all_dfs)

        return pd.DataFrame()

    def _get_feature_metadata(
        self,
        model: management_types_pb2.Model,
        featureset_map: Dict[str, management_types_pb2.FeatureSet],
    ):
        try:
            self._query = Query.form_query(
                model=model,
                metadata=True,
                all_featuresets=featureset_map,
            )
        except Exception as e:
            raise Exception(f"Unable to form query for model {model.name}: {str(e)}")

        (
            input_features_dict_lst,
            output_features_dict_lst,
        ) = self._query.get_model_version_io(include_id=True)
        if not input_features_dict_lst or not output_features_dict_lst:
            raise Exception(
                f"Input features {input_features_dict_lst} or output features {output_features_dict_lst} invalid."
            )

        if len(output_features_dict_lst) > 1:
            raise ValueError(
                f"Currently cannot handle multiple outputs: {output_features_dict_lst}"
            )

        feature_names = set()
        for feature in input_features_dict_lst:
            if feature["name"] not in feature_names:
                self._input_features.append(types.QueryFeature(**feature))
                feature_names.add(feature["name"])

        self._target = types.QueryFeature(**output_features_dict_lst[0])
        logging.info(f"Feature {output_features_dict_lst[0]}, target {self._target}")



import json
import tempfile
import traceback
from continual.python.sdk.datasets import Dataset


from continual.python.sdk.extensions.models.base_model_pipeline import BaseModelPipeline
from continual.python.sdk.client import Client
from continual.services.compute_services.utils.compute_logger import (
    get_logger,
)
from continual.services.compute_services.pipelines.pipeline_utils import get_pipeline

LOG_FILE_NAME = "training.runner.log.txt"


class TrainingPipelineRunner:
    def __init__(
        self,
        client: Client,
        model_version_name: str,
        logging_level: str = "INFO",
    ):

        self.client = client
        self.config = client.model_versions.get_train_config(
            model_version_name=model_version_name
        )
        self.logging_level = logging_level

        # Construct utils
        self.logger = get_logger(__name__, LOG_FILE_NAME, logging_level)
        self.logger.info(f"Starting TrainingPipeline runner for {model_version_name} ")

        # Construct arguments
        self.model_version = self.config.model_version
        self.model = self.config.model.to_proto()
        self.featureset_map = {}
        for fs in self.config.feature_sets:
            self.logger.info(f"feature set joined in: {fs}")
            self.featureset_map[fs.name] = fs.to_proto()

        if not self.config.feature_sets:
            self.logger.info(
                f"No feature sets found to join in for model version: {model_version_name}."
            )

        self.data_store = self.config.data_store.to_proto()
        self.tag = self.config.tag
        self.evaluation_model_version = self.config.evaluation_model_version
        self.evaluation_model_signature = self.config.evaluation_model_signature
        self.output_dir = tempfile.mkdtemp()
        self.fit_config = {}
        if self.model_version.fit_config != "":
            self.fit_config = json.loads(self.model_version.fit_config)

    def run(self):
        """
        Constructs pipeline
        Constructs train args
        Constructs promote args
        Runs Train
        Runs Promote
        Logs errors
        """
        # Collect data
        try:
            train_data = Dataset(
                model=self.model,
                featureset_map=self.featureset_map,
                data_store=self.data_store,
            )
        except Exception as e:
            stack_trace = traceback.format_exc()
            self._set_errors("Error constructing dataset", e, stack_trace)
            raise

        # Get and init pipeline
        try:
            pipeline: BaseModelPipeline = get_pipeline(
                client=self.client, pipeline_id=self.model.schema.pipeline.name
            )()
        except Exception as e:
            stack_trace = traceback.format_exc()
            self._set_errors(f"Exception getting pipeline", e, stack_trace)
            raise

        # Training
        self._train_args = dict(
            client=self.client,
            model_version_name=self.model_version.name,
            train_data=train_data,
            output_dir=self.output_dir,
            fit_config=self.fit_config,
            evaluation_model_version_name=self.evaluation_model_version,
            evaluation_model_signature=self.evaluation_model_signature,
            model=self.model,
            tag=self.tag,
            data_store=self.data_store,
            log_level=self.logging_level,
        )

        try:
            pipeline.train(**self._train_args)
        except Exception as e:
            stack_trace = traceback.format_exc()
            self._set_errors(f"Exception during train", e, stack_trace)
            raise

        # Successful train and promote
        self._set_errors(clear=True)

    def _set_errors(
        self, message=None, e: Exception = None, stack_trace: str = None, clear=False
    ):
        """
        Make sure errors are recorded in the logs as well as the
        model version

        Parameters
        ---------
        message: str
            Description of what pipeline step the exception occured in
        e: Exception
            The exception itself
        """

        if clear:
            self.client.logger.log_field("error_message", "")
            self.client.logger.log_field("stack_trace", "")
        else:
            error_message = f"{message}:\n {e}"
            self.logger.error(f"{message}:\n {e} : {stack_trace}")
            self.client.logger.log_field("error_message", error_message)
            self.client.logger.log_field("stack_trace", stack_trace)



import logging
import os
import traceback
import logging


from continual.python.sdk.exceptions import InternalError, InvalidArgumentError
from continual.python.blob.store import BlobStore
from continual.python.utils.utils import read_csv
from tableschema import Table


def is_column_categorical(series):
    unique_count = series.nunique(dropna=False)
    total_count = len(series)
    return total_count > unique_count


class FeatureSourceCloudStore:
    def __init__(self, id, filename):

        self.filename = filename
        self._id = id
        if filename is not None and len(filename) > 0:
            if not filename.startswith("gs://"):
                raise InvalidArgumentError(
                    "only google cloud store source supported", None
                )
            self.blob_store = BlobStore(filename)
        else:
            raise InvalidArgumentError("source is not file type", None)

    def _build_object_list(self):
        objects = []
        if self.filename is not None and len(self.filename) > 0:
            objects.append(self.filename)
        return objects

    def fetch_all(self):
        output = "/tmp/" + self._id
        self.file_path = self.blob_store.download_file(self.filename, outpath=output)
        return read_csv(self.file_path)

    def fetch_data(self, page_size=10000):
        objects = self._build_object_list()

        class BlobIterator:
            def __init__(self, id, blob_store, objects):
                self._blob_store = blob_store
                self._id = id
                self._objects = objects
                self._current_index = 0
                self.chunker = None
                self.file_path = None

            def __iter__(self):
                self._current_pointer = 0
                self.chunker = None
                return self

            def __next__(self):
                if self.chunker is not None:
                    return next(self.chunker)

                if (
                    self._current_index >= len(self._objects)
                    and self.file_path is not None
                ):
                    os.remove(self.file_path)
                    raise StopIteration

                output = "/tmp/" + self._id
                self.file_path = self._blob_store.download_file(
                    self._objects[self._current_index], outpath=output
                )
                logging.info("Downloaded " + self.file_path)
                self.chunker = read_csv(self.file_path, chunksize=page_size)

                self._current_index = self._current_index + 1
                return next(self.chunker)

        return BlobIterator(self._id, self.blob_store, objects)

    def infer_schema(self, filename, sample_row_count=100):

        # Download the file from the blob store.
        self.blob_store = BlobStore(filename)
        file_path = self.blob_store.download_file(filename, outpath="/tmp")
        logging.info("Downloaded " + file_path)

        inferred_schema = {"metadata": {}, "features": {}, "entities": {}, "models": {}}
        try:
            table = Table(file_path)
            table.infer()
            for f in table.schema.descriptor["fields"]:
                col = f["name"].strip()
                if col == "":
                    continue
                elif col.lower() == "id":
                    inferred_schema["metadata"][col] = "TEXT"
                    continue

                if len(col) > 63:
                    print("SKIPPING ", col)
                    continue

                if f["type"] == "boolean":
                    inferred_schema["features"][col] = "BOOLEAN"
                elif f["type"] == "number" or f["type"] == "integer":
                    inferred_schema["features"][col] = "NUMBER"
                elif f["type"] == "string":
                    inferred_schema["features"][col] = "TEXT"

            self.chunker = read_csv(file_path, chunksize=sample_row_count or 1)
            df = next(self.chunker)

            # Determine categorical columns.
            for k in inferred_schema["features"]:
                if inferred_schema["features"][k] != "TEXT":
                    continue
                if k in df.columns and is_column_categorical(df[k]):
                    inferred_schema["features"][k] = "CATEGORICAL"

            return inferred_schema, df

        except Exception:
            traceback.print_exc()
            raise InternalError("could not read source")

    def close(self):
        self._blob_store = None



from abc import ABC, abstractclassmethod
import pandas as pd

from continual.python.utils.utils import parse_model_name
from .featurestore import get_featurestore

from jinja2 import Template
from datetime import datetime
from urllib.parse import urlparse, parse_qs

PREDICTION_TABLE_NAME = "predictions"

# running into "Compilation memory exhausted" errors so setting lower
# TODO: optimize this size
WRITE_CHUNK_SIZE = 1000

# Based on Snowflake limit of 1M
MAX_QUERY_LEN = 900 * 1024 * 1024

insert_template = Template(
    """INSERT INTO  {{param.project}}.{{param.table_name}} ({% for r in param.insert_list %}{%if loop.index0 != 0%},{% endif %}{{r}}{% endfor %})
        SELECT {% for r in param.columns %}{%if loop.index0 != 0%},{% endif %}{{r}}{% endfor %} 
        FROM {{param.project}}.{{param.temp_table}}"""
)


def parse_connection_url(connection_url):
    upass = urlparse(connection_url)
    qpass = parse_qs(upass.query)
    upass_splits = upass.netloc.split("@")
    username = upass_splits[0].split(":")[0]
    password = upass_splits[0].split(":")[1]
    account = upass_splits[1]
    database = upass.path.strip("/")
    warehouse = qpass.get("warehouse", None)
    if warehouse is not None:
        warehouse = warehouse[0]
    role = qpass.get("role", None)
    if role is not None:
        role = role[0]

    return account, username, password, database, warehouse, role


def chunk_it(start, end, step=WRITE_CHUNK_SIZE):
    if start >= end:
        return None

    if start + step > end:
        return (start, end)

    return (start, start + step)


class PredictionStore(ABC):
    def __init__(self, datastore):
        # db_url = "postgresql://continual:continual@localhost:5432/continual"
        self.datastore = datastore
        self.engine = None
        self.connection = None

    @staticmethod
    def getStore(
        datastore,
        cfg: dict = None,
    ):
        if datastore is None:
            return None

        return get_featurestore(data_store=datastore, cfg=cfg)

    def parse_batchPrediction_job_name(self, name):
        splits = name.split("/")
        project_name = "/".join(splits[:2])
        model_name = "/".join(splits[:4])
        return project_name, model_name

    def get_model_prediction_table_name(self, model):
        project_id, model_id = parse_model_name(model.name)
        project_id = project_id.replace("@", "_")
        return project_id, "model_{}_predictions_history".format(model_id)

    def _process_dataframe(
        self, df, id_column_name, ts_column_name, name, model_version
    ):
        for col in df.columns:
            if col.startswith("Unnamed:"):
                df = df.drop(col, axis=1)

        df.columns = map(str.lower, df.columns)
        # new query object should support better implementation
        for col in df.columns:
            if col.endswith(f":{id_column_name}"):
                df = df.rename(columns={col: id_column_name})
            if col.endswith(f":{ts_column_name}"):
                df = df.rename(columns={col: ts_column_name})

        all_times = list(df.select_dtypes(include=["datetimetz", "datetime64"]).columns)
        for t in all_times:
            df[t] = df[t].apply(lambda x: x.isoformat())

        df["batch_prediction"] = name
        df["model_version"] = model_version

        df = df.where(pd.notnull(df), None)
        return df

    @abstractclassmethod
    def write_prediction(
        self,
        df: pd.DataFrame,
        id_column_name: str,
        ts_column_name: str,
        model_query: any,
        output_name: str,
        job_name: str,
        job_start_time: datetime,
        model: any,
        model_version: str,
        time_index_dtype,
    ):
        pass

    def close(self):
        pass



import logging
import os
import sys
from typing import List

from continual.services.compute_services.utils.compute_logger import (
    get_logger,
    configure_logging,
    StandardOutputSplicer,
    StandardErrorSplicer,
)

MASTER_LOG_PATH = os.path.join(os.path.dirname(__file__), "master.test.log")


def test_configure_logger():
    cleanup()

    # First check that warnings show up and infos dont
    configure_logging(MASTER_LOG_PATH)
    logging.info("MASTER_LOG_TEST")
    logging.info("MASTER_LOG_TEST")
    logging.info("MASTER_LOG_TEST")
    logging.debug("MASTER_LOG_TEST")
    logging.debug("MASTER_LOG_TEST")

    with open(MASTER_LOG_PATH, "r") as f:
        contents = f.read()
    assert contents.count("MASTER_LOG_TEST") == 3

    # Now get a random logger from logging and perform same test
    logger = logging.getLogger("test_logger")

    logger.info("MASTER_LOG_TEST")
    logger.debug("MASTER_LOG_TEST")
    with open(MASTER_LOG_PATH, "r") as f:
        contents = f.read()
    assert contents.count("MASTER_LOG_TEST") == 4
    cleanup()


def test_redirect_std():
    cleanup()
    configure_logging(MASTER_LOG_PATH)
    with StandardOutputSplicer(MASTER_LOG_PATH), StandardErrorSplicer(MASTER_LOG_PATH):
        logging.info("LOG TEST")
        print("STDOUT TEST")
        print("STDERR TEST", file=sys.stderr)

    with open(MASTER_LOG_PATH, "r") as f:
        contents = f.read()

    assert contents.find("LOG TEST") != -1
    assert contents.find("STDOUT TEST") != -1
    assert contents.find("STDERR TEST") != -1
    cleanup()


def test_get_logger():
    log_files = [
        os.path.join(os.path.dirname(__file__), filename)
        for filename in [f"logger.{i+1}.log" for i in range(2)]
    ]
    cleanup(log_files)
    configure_logging(MASTER_LOG_PATH)
    # Create 1 logger

    logger1 = get_logger("logger1", log_files[0], log_level=logging.DEBUG)
    logger1.debug("DEBUG1")
    logger2 = get_logger("logger2", log_files[1], log_level=logging.DEBUG)
    logger2.debug("DEBUG2")

    # Check files
    with open(log_files[0], "r") as f1, open(log_files[1], "r") as f2, open(
        MASTER_LOG_PATH, "r"
    ) as fm:
        log1_contents = f1.read()
        log2_contents = f2.read()
        # Ensure master log has both
        master_file_contents = fm.read()

    assert (
        master_file_contents.find("DEBUG1") != -1
        and master_file_contents.find("DEBUG2") != -1
    )
    assert log1_contents.find("DEBUG1") != -1
    assert log2_contents.find("DEBUG2") != -1

    cleanup(log_files)


def cleanup(logs: List[str] = []):
    if os.path.exists(MASTER_LOG_PATH):
        os.remove(MASTER_LOG_PATH)
    for log in logs:
        if os.path.exists(log):
            os.remove(log)



from typing import Callable, List, TypeVar, Generic, Tuple

T = TypeVar("T")
NextPageFunction = Callable[[str], Tuple[List[T], str]]


class Pager(Generic[T]):
    """Pager."""

    next_page: NextPageFunction

    def __init__(self, next_page: NextPageFunction):
        self.next_page = next_page

    def __iter__(self):
        return PagerInterator(self)

    def to_list(self) -> List[T]:
        """Convert iterator to list."""
        return list(self)


class PagerInterator(Generic[T]):
    """Creates an iterator form a list operation."""

    def __init__(self, source: Pager[T]):
        self.source = source
        self.next_page_token = ""
        self.current_pointer = 0
        self.current_page = None

    def __next__(self) -> T:

        if self.current_page is None or (
            self.current_pointer == len(self.current_page)
            and self.next_page_token != ""
        ):
            self.current_page, self.next_page_token = self.source.next_page(
                self.next_page_token
            )
            self.current_pointer = 0

        if self.current_pointer < len(self.current_page):
            self.current_pointer += 1
            return self.current_page[self.current_pointer - 1]

        raise StopIteration



from __future__ import annotations
from typing import List, Optional, Tuple
from google.protobuf import field_mask_pb2
from continual.python.utils import utils
from continual.rpc.management.v1 import management_pb2
from continual.rpc.management.v1 import types
from continual.python.sdk import client
from continual.python.sdk.resource import Resource
from continual.python.sdk.manager import Manager
from continual.python.sdk.iterators import Pager
from continual.python.sdk.featuresets import FeatureSetManager
from continual.python.sdk.models import ModelManager
from continual.python.sdk.extensions import ExtensionManager
from continual.python.sdk.changes import ChangeManager
from continual.python.sdk.events import EventManager
import requests


class ProjectEnvironmentManager:
    """Manages projects environments."""

    manager: EnvironmentManager
    parent: str = ""
    client: client.Client

    def __init__(self, parent: str, client: client.Client):
        self.parent = parent
        self.client = client
        self.manager = EnvironmentManager(parent=self.parent, client=self.client)

    def get(self, id: str) -> Environment:
        """Get an environment.

        Arguments:
            id: The environment ID.

        Returns:
            The environment with the provided ID.
        """
        return self.manager.get(id)

    def list(self, page_size: Optional[int] = None) -> List[Environment]:
        """List environments.

        Arguments:
            page_size: Number of elements.

        Returns:
            A list of project environments.
        """
        return self.manager.list(page_size)

    def list_all(self) -> Pager[Environment]:
        """List all environments.

        Returns:
            An iterator of all enviroments.
        """
        return self.manager.list_all()

    def create(
        self,
        id: str,
        data_store: Optional[types.DataStore] = None,
        source: str = "",
    ) -> Environment:
        """Create a new Environment.

        Creates a new environment within project.  The environment ID must
        be unique within the project.

        Arguments:
            id: Environment ID.
            data_store: Data store configuration.
        Returns:
            A new Environment
        """
        return self.manager.create(
            id=id,
            source=source,
            data_store=data_store,
        )


class EnvironmentManager(Manager):
    """Manages environment resources."""

    name_pattern: str = "projects/{project}"

    def __init__(self, client: client.Client, parent: str = "") -> None:
        self.client = client
        if "@" in parent:
            self.parent = parent.split("@")[0]
        else:
            self.parent = parent

    def name(self, id: str, parent: Optional[str] = None):
        if "/" in id:
            # Don't allow names to override manager parent config since this is confusing
            # and is typically a bug in the user code.
            if parent is not None and parent != "" and not id.startswith(parent):
                raise ValueError(f"Resource {id} not a child of {parent}.")
            return id

        name_str = self.parent or ""
        name_str += "@" + id
        return name_str

    def get(self, id: str) -> Environment:
        """Get environment.

        Arguments:
            id: environment name or id.

        Returns
            A environment.
        """
        req = management_pb2.GetEnvironmentRequest(name=self.name(id))
        resp = self.client._management.GetEnvironment(req)
        return Environment.from_proto(resp, client=self.client)

    def list(
        self, page_size: Optional[int] = None, filters: List[str] = None
    ) -> List[Environment]:
        """List environments.

        Arguments:
            page_size: Number of items to return.

        Returns:
            A list of environments.
        """
        req = management_pb2.ListEnvironmentsRequest(
            parent=self.parent, page_size=page_size, filters=filters
        )
        resp = self.client._management.ListEnvironments(req)
        return [
            Environment.from_proto(x, client=self.client) for x in resp.environments
        ]

    def list_all(self) -> Pager[Environment]:
        """List all environments.

        Pages through all environments using an iterator.

        Returns:
            A iterator of all environments.
        """

        def next_page(next_page_token):
            req = management_pb2.ListEnvironmentsRequest(
                parent=self.parent, page_token=next_page_token
            )
            resp = self.client._management.ListEnvironments(req)
            return (
                [
                    Environment.from_proto(x, client=self.client)
                    for x in resp.environments
                ],
                resp.next_page_token,
            )

        return Pager(next_page)

    def create(
        self,
        id: str,
        source: str = "",
        data_store: Optional[types.DataStore] = None,
        scheduling_enabled: bool = False,
    ) -> Environment:
        """Create an environment.

        New environments are identified by a unique ID within
        their parent project.

        Arguments:
            id: Environment ID.
            data_store: Data store configuration.

        Returns:
            A new environment.
        """
        req = management_pb2.CreateEnvironmentRequest(
            parent=self.parent,
            environment_id=id,
            source=source,
            data_store=data_store,
            scheduling_enabled=scheduling_enabled,
        )
        resp = self.client._management.CreateEnvironment(req)
        return Environment.from_proto(resp, client=self.client)

    def delete(self, id: str, delete_schema: bool = False) -> None:
        """Delete an Environment.

        Arguments:
            id: Environment name or id.
        """

        req = management_pb2.DeleteEnvironmentRequest(
            name=self.name(id), delete_schema=delete_schema
        )
        self.client._management.DeleteEnvironment(req)

    def update(
        self,
        id: str,
        scheduling_enabled: Optional[bool] = None,
        data_store: Optional[types.DataStore] = None,
    ) -> Environment:
        """Update Environment.

        Arguments:
            id: Environment ID.
            scheduling_enabled: Whether to enabled scheduled jobs in the Environment.
            data_store: Environment's data store.
        Returns:
            Updated Environment.
        """
        paths = []

        env = Environment(
            name=self.name(id),
        )

        if data_store is not None:
            env.data_store = data_store
            paths.extend(utils.generate_datastore_field_paths(data_store))

        if scheduling_enabled is not None:
            env.scheduling_enabled = scheduling_enabled
            paths.append("scheduling_enabled")

        req = management_pb2.UpdateEnvironmentRequest(
            update_mask=field_mask_pb2.FieldMask(paths=paths),
            environment=env.to_proto(),
        )
        resp = self.client._management.UpdateEnvironment(req)
        return Environment.from_proto(resp, client=self.client)


class Environment(Resource, types.Environment):
    """Environment resource."""

    name_pattern: str = "projects/{project}"
    manager: EnvironmentManager

    feature_sets: FeatureSetManager
    """Feature set manager."""

    models: ModelManager
    """Model manager."""

    extensions: ExtensionManager
    """Extension manager."""

    changes: ChangeManager
    """Change manager."""

    events: EventManager
    """Event manager."""

    def _init(self):
        self.manager = EnvironmentManager(parent=self.parent, client=self.client)
        self.feature_sets = FeatureSetManager(parent=self.name, client=self.client)
        self.models = ModelManager(parent=self.name, client=self.client)
        self.extensions = ExtensionManager(parent=self.name, client=self.client)
        self.changes = ChangeManager(parent=self.name, client=self.client)
        self.events = EventManager(parent=self.name, client=self.client)

    def delete(self, delete_schema: bool = False) -> None:
        """Delete environment."""
        self.manager.delete(self.name, delete_schema=delete_schema)

    def update(
        self,
        scheduling_enabled: Optional[bool] = None,
        data_store: Optional[types.DataStore] = None,
    ) -> Environment:
        """Update environment.

        Arguments:
            scheduling_enabled:  Enable scheduled jobs within the Environment.

        Returns:
            Updated environment.
        """
        return self.manager.update(
            self.name, scheduling_enabled=scheduling_enabled, data_store=data_store
        )

    def _create_csv_signed_url(self, file_name: str) -> str:
        """Get a temporary google signed url.

        Arguments:
            file_name: name of the file to get a signed url for.
        """
        req = management_pb2.CreateFeatureSetCSVSignedURLRequest(
            parent=self.name, file_name=file_name
        )
        resp = self.client._management.CreateFeatureSetCSVSignedURL(req)
        return resp.signed_url, resp.bucket_path

    def upload_csv_file(self, file_name: str) -> str:
        """upload a csv file to a continual bucket.

        Arguments:
            file_name: name of the file to upload.
        """
        signed_url, bucket_path = self._create_csv_signed_url(file_name)
        TYPE = "text/csv"
        # TODO: chunk up the file if needed
        resp = requests.put(
            signed_url,
            open(file_name).read().encode("utf-8"),
            headers={"Content-Type": TYPE},
        )
        if resp.text != "":
            raise Exception("could not upload file {}: {}".format(file_name, resp.text))

        return bucket_path

    def seed(self, file_name: str, table_name: str = None) -> str:
        """upload a csv file to a continual bucket.

        Arguments:
            file_name: name of the file to upload.
        """
        if not file_name.startswith("gs://"):
            signed_url, bucket_path = self._create_csv_signed_url(file_name)
            TYPE = "text/csv"
            # TODO: chunk up the file if needed
            resp = requests.put(
                signed_url,
                open(file_name).read().encode("utf-8"),
                headers={"Content-Type": TYPE},
            )
            if resp.text != "":
                raise Exception(
                    "could not upload file {}: {}".format(file_name, resp.text)
                )
        else:
            bucket_path = file_name

        req = management_pb2.SeedRequest(
            project=self.name, file_path=bucket_path, table_name=table_name
        )
        resp = self.client._management.Seed(req)
        return resp.schema_name, resp.table_name

    def export(self) -> Tuple[dict, dict, dict]:
        req = management_pb2.ExportEnvironmentRequest(name=self.name)
        resp = self.client._management.ExportEnvironment(req)
        return resp.feature_sets, resp.models, resp.extensions

    """
    def get_schema_graph(self) -> str:
        "" "The a JSON graph of current schema within the project.

        Arguments:
            None
        " ""

        req = management_pb2.GetSchemaGraphRequest(name=self.name)
        resp = self.client._management.GetSchemaGraph(req)
        return resp.graph_json
    """



import os
import logging
import urllib.parse
from urllib.parse import urlparse
from urllib3.exceptions import ProtocolError
from requests.exceptions import ChunkedEncodingError
from google.cloud import storage
from google.api_core import retry

logger = logging.getLogger(__name__)


class GoogleCloudStorage(object):
    def __init__(self, resource_url, **kwargs):
        self.resource_url = resource_url
        self._register_gcs_scheme()

        # GCS has a download timeout of 60s. There doesn't seem to be a
        # way to alter it but we can download in chunks.
        storage.blob._DEFAULT_CHUNKSIZE = 100 * 1024 * 1024  # 100M
        storage.blob._MAX_MULTIPART_SIZE = 100 * 1024 * 1024  # 100M

        self.bucket_name = kwargs.get("bucket_name", self._parse_bucket_name())

        if "project" in kwargs or "credentials" in kwargs:
            self.client = storage.Client(
                project=kwargs.get("project"), credentials=kwargs.get("credentials")
            )
        else:
            self.client = storage.Client()

    @staticmethod
    def is_valid_url(resource_url):
        if not resource_url.startswith("gs://"):
            logger.warning("resource URL does not start with gs://")
            return False
        elif not resource_url[len("gs://") :]:
            logger.warning("resource URL does not contain bucket name after gs://")
            return False

        return True

    def list_objects(self):
        blobs = self.client.list_blobs(
            self.bucket_name, prefix=self._get_path_from_url(self.resource_url)
        )
        objs = []
        for b in blobs:
            if b.name.endswith("/"):
                continue
            objs.append(b.name)
        return objs

    def _register_gcs_scheme(self):
        def register_scheme(scheme):
            for method in filter(lambda s: s.startswith("uses_"), dir(urllib.parse)):
                getattr(urllib.parse, method).append(scheme)

        register_scheme("gs")

    def _parse_bucket_name(self):
        parts = urlparse(self.resource_url)._asdict()
        bucket = parts.get("netloc")  # gs://bucket => bucket
        if not bucket:
            raise ValueError(
                "cannot extract bucket name from resource URL, found {}".format(bucket)
            )
        return bucket

    def _get_or_create_bucket(self, bucket_name=""):
        bucket_name = bucket_name or self.bucket_name
        try:
            bucket = self.client.get_bucket(bucket_name)
            logger.info("bucket already exists with name {}...".format(bucket_name))
            return bucket
        except Exception as e:
            logger.info(
                f"Encountered exception [{e}] for bucket [{bucket_name}], will try to create it"
            )
            bucket = self.client.create_bucket(bucket_name)
            return bucket

    def _get_path_from_url(self, url: str) -> str:
        parts = urlparse(url)._asdict()
        path = str(parts.get("path"))
        if path and path[0] == "/":
            path = path[1:]

        return path

    def upload_file(self, filepath, prefix="", bucket_name="", bucket=None):
        bucket_name = bucket_name or self.bucket_name
        if not bucket:
            bucket = self._get_or_create_bucket(bucket_name)

        filename = os.path.split(filepath)[-1]
        blob_name = os.path.join(prefix, filename)
        blob = bucket.blob(blob_name)

        logger.info("Uploading ({}) to ({})".format(filepath, blob_name))
        blob.upload_from_filename(filepath)
        logger.info("Uploaded ({}) to ({})".format(filepath, blob_name))

    def upload_dir(self, dirpath: str, prefix: str = "", bucket_name: str = "") -> None:
        logger.info("Uploading files recursively from directory: ({})".format(dirpath))
        bucket_name = bucket_name or self.bucket_name
        bucket = self._get_or_create_bucket(bucket_name)

        for root, _dirs, files in os.walk(dirpath):
            for file in files:
                nested_prefix = os.path.relpath(root, dirpath)
                if nested_prefix == ".":
                    nested_prefix = ""

                nested_prefix = os.path.join(prefix, nested_prefix)

                self.upload_file(
                    os.path.join(root, file),
                    prefix=nested_prefix,
                    bucket_name=bucket_name,
                    bucket=bucket,
                )
        return

    def download_file(
        self, blobpath, outpath="", bucket_name="", bucket=None, base_name=""
    ):
        bucket_name = bucket_name or self.bucket_name
        if not bucket:
            bucket = self._get_or_create_bucket(bucket_name)

        if self.is_valid_url(blobpath):
            blobpath = self._get_path_from_url(blobpath)

        blob = bucket.get_blob(blobpath)
        if not blob:
            raise Exception("could not find blob to download: {}".format(blobpath))

        blobname = blob.name[len(base_name) + 1 :]
        if base_name == "":
            blobname = blob.name.split("/")[-1]
        dl_path = os.path.join(outpath, blobname)
        path = "/"
        path = path.join(dl_path.split("/")[:-1])
        if not os.path.isdir(path):
            print("making path: " + path)
            os.makedirs(path)

        retrier = retry.Retry(
            predicate=retry.if_exception_type(
                ConnectionResetError,
                ProtocolError,
                ChunkedEncodingError,
            ),
        )
        retrier(blob.download_to_filename)(dl_path)

        return dl_path

    def download_as_string(self, blobpath=None):
        bucket = self._get_or_create_bucket(None)
        if blobpath is not None and self.is_valid_url(blobpath):
            blobpath = self._get_path_from_url(blobpath)
        else:
            blobpath = self._get_path_from_url(self.resource_url)

        if blobpath is None:
            raise ValueError(f"No blob name in path [{blobpath}]")

        blob = bucket.get_blob(blobpath)
        if not blob:
            raise ValueError("could not find blob to download: [{}]".format(blobpath))

        ret_str = blob.download_as_string()
        return ret_str

    def download_dir(
        self, dirpath: str, outpath: str = "", bucket_name: str = ""
    ) -> str:
        bucket_name = bucket_name or self.bucket_name
        bucket = self._get_or_create_bucket(bucket_name)

        if self.is_valid_url(dirpath):
            dirpath = self._get_path_from_url(dirpath)

        blobs = bucket.list_blobs(prefix=dirpath)

        for blob in blobs:

            blobname = blob.name
            self.download_file(
                blobname,
                outpath=outpath,
                bucket_name=bucket_name,
                bucket=bucket,
                base_name=dirpath,
            )

        return os.path.join(outpath, dirpath)



import json
import grpc


def get_management_channel(
    endpoint, service="continual.rpc.management.v1.ManagementAPI", options=None
):
    options = options or []
    service_config = json.dumps(
        {
            "methodConfig": [
                {
                    "name": [
                        {
                            "service": service
                            or "continual.rpc.management.v1.ManagementAPI"
                        }
                    ],
                    "retryPolicy": {
                        "maxAttempts": 5,
                        "initialBackoff": "1s",
                        "maxBackoff": "30s",
                        "backoffMultiplier": 2,
                        "retryableStatusCodes": ["UNAVAILABLE", "RESOURCE_EXHAUSTED"],
                    },
                }
            ]
        }
    )
    options.extend(
        [
            ("grpc.max_receive_message_length", 100 * 1024 * 1024),
            ("grpc.max_send_message_length", 100 * 1024 * 1024),
            ("grpc.service_config", service_config),
        ]
    )

    if endpoint == "https://api.continual.ai":
        credentials = grpc.ssl_channel_credentials()
        return grpc.secure_channel("api.continual.ai:443", credentials, options)
    elif endpoint == "https://api-benchmark.continual.ai":
        credentials = grpc.ssl_channel_credentials()
        return grpc.secure_channel(
            "api-benchmark.continual.ai:443", credentials, options
        )
    elif endpoint == "http://localhost" or endpoint == "http://127.0.0.1":
        host = "127.0.0.1:30003"
        return grpc.insecure_channel(host, options)
    elif endpoint == "http://host.docker.internal":
        host = "host.docker.internal:30003"
        return grpc.insecure_channel(host, options)
    elif endpoint == "http://cluster":
        host = "management:3001"
        return grpc.insecure_channel(host, options)
    else:
        return grpc.insecure_channel(endpoint, options)



import logging

# get_context_sesssion gets the session token from the auth bearer
# currently this is used directly in the grpc servers to obtain a token to use
# for subsequent calls back to the management service.  We may want to move this
# or in addition, add it to an interceptor for better authorization checks.
def get_context_sesssion(context):
    try:
        metadict = dict(context.invocation_metadata())
        logging.info(
            "context.invocation_metadata(): {}".format(context.invocation_metadata())
        )
        auth = metadict["authorization"]
        logging.info("auth: {}".format(auth))
        token = auth.split(" ")[1]
        logging.info("token: {}".format(token))
        if token.startswith("sessions/"):
            return token
        logging.error(f"invalid token: {token}")
        raise Exception(f"invalid token: {token}")

    except Exception as e:
        logging.error(f"error getting session: {e}")
        raise Exception(f"error getting session: {e}")



from continual.rpc.management.v1 import management_types_pb2
import yaml


def get_type_enum(typename):
    if typename is None:
        return None

    return management_types_pb2._FIELDTYPE.values_by_name[typename.upper()].number


def load_schemas(file_name):

    sc = None
    schema_text = None
    with open(file_name, "r") as inp_file:
        schema_text = inp_file.read()

    with open(file_name, "r") as inp_file:
        sc = yaml.load(inp_file, Loader=yaml.FullLoader)
        # full_name = "projects/" + project + '/featureSets/' + sc['name'] + "@1"

    if sc is None or "type" not in sc:
        return None

    if sc["type"] == "FeatureSet":
        schema = management_types_pb2.FeatureSetSchema(
            type=sc["type"],
            name=sc["name"],
            entity=sc["entity"],
            table=sc.get("table", None),
            description=sc.get("description", None),
            url=sc.get("url", None),
            owners=sc.get("owners", None),
            documentation=sc.get("documentation", None),
            query=sc.get("query", None),
            exclude_columns=sc.get("exclude_columns", None),
            columns=[],
        )
    elif sc["type"] == "Model":
        schema = management_types_pb2.ModelSchema(
            type=sc["type"],
            name=sc["name"],
            table=sc.get("table", None),
            description=sc.get("description", None),
            url=sc.get("url", None),
            owners=sc.get("owners", None),
            documentation=sc.get("documentation", None),
            query=sc.get("query", None),
            exclude_columns=sc.get("exclude_columns", None),
            columns=[],
            target=sc.get("target", None),
        )

        if "train" in sc:
            if "schedule" in sc["train"]:
                schema.train.schedule = sc["train"]["schedule"]
            if "metric" in sc["train"]:
                schema.train.metric = sc["train"]["metric"]
            if "optimization" in sc["train"]:
                schema.train.optimization = sc["train"]["optimization"]
            if "advanced" in sc["train"]:
                schema.train.advanced = sc["train"]["advanced"]

        if "promote" in sc and "policy" in sc["promote"]:
            schema.promote.policy = sc["promote"]["policy"]

        if "predict" in sc:
            if "schedule" in sc["predict"]:
                schema.predict.schedule = sc["predict"]["schedule"]
            if "incremental" in sc["predict"]:
                schema.predict.incrememtal = sc["predict"]["incremental"]

    else:
        raise Exception("Unknow schema type " + sc["type"])

    if "columns" in sc:
        column_configs = []
        columns = sc["columns"]
        for entry in columns:
            mc = management_types_pb2.ColumnConfig(
                name=entry["name"].lower(),
                description=entry.get("description", None),
                type=get_type_enum(entry.get("type", None)),
                entity=entry.get("entity", None),
            )
            column_configs.append(mc)
        schema.columns.extend(column_configs)

    return schema, schema_text



#!/usr/bin/env python3

import typer

from continual.python.cli import utils
from continual.python.sdk.client import Client
from continual.python.cli import projects

from rich.console import Console
from typing import List
from pathlib import Path
from enum import Enum

app = typer.Typer(help="Manage environments.")


class SchemaType(str, Enum):
    FeatureSet = "FeatureSet"
    Model = "Model"


def format_environment_data(e, zipped=False):
    create_time = e.create_time.replace(microsecond=0)
    update_time = e.update_time.replace(microsecond=0)
    fs_type = e.data_store.type
    (db_schema, database) = utils.get_datastore_database_and_schema(e.data_store)

    c = Client()

    environment_id = e.id
    current_project = utils.get_project(c.config.project)
    current_environment = utils.get_environment(current_project, c.config.environment)

    if (
        "@" not in environment_id
        and current_project.endswith(environment_id)
        and (current_environment in ("prod", "production") or not current_environment)
    ):
        environment_id = f"{environment_id} (active)"
    elif e.name == utils.get_environment_name(current_project, current_environment):
        environment_id = f"{environment_id} (active)"

    if zipped:
        data = [
            environment_id,
            e.source,
            fs_type,
            database,
            db_schema,
            e.scheduling_enabled,
            e.summary.feature_set_count,
            e.summary.feature_count,
            e.summary.model_count,
            e.summary.model_version_count,
            e.summary.prediction_count,
            create_time,
            update_time,
        ]
        headers = [
            "ID",
            "Source",
            "Data Warehouse",
            "Database",
            "Database Schema",
            "Scheduling Enabled",
            "Feature Sets",
            "Features",
            "Models",
            "Model Versions",
            "Predictions",
            "Created",
            "Updated",
        ]
        return tuple(
            [x[0], x[1]] for x in (zip(headers, data))
        )  # for some reason list(zip) causes issues, so ...
    else:
        data = [
            environment_id,
            fs_type,
            db_schema,
            e.scheduling_enabled,
            e.summary.feature_set_count,
            e.summary.feature_count,
            e.summary.model_count,
            e.summary.model_version_count,
            e.summary.prediction_count,
            create_time,
            # update_time,
        ]
        headers = [
            "ID",
            "Data Warehouse",
            "Database Schema",
            "Scheduling Enabled",
            "Feature Sets",
            "Features",
            "Models",
            "Model Versions",
            "Predictions",
            "Created",
            # "Updated",
        ]
        return (data, headers)


# use callback to run list command if nothing is passed in
@app.callback(invoke_without_command=True)
def default(ctx: typer.Context):
    if ctx.invoked_subcommand is not None:
        return
    else:
        list(n=30, project=None, filters=[], style=None)


@app.command("list")
@utils.exit_on_error
def list(
    project: str = typer.Option(None, help="Project ID."),
    n: int = typer.Option(30, "--num", "-n", help="Number of records to show."),
    filters: List[str] = typer.Option([], "--filter", "-f", help="List of filters."),
    style: utils.ContinualStyle = typer.Option(None, help="Color to use for list."),
):
    """List environments."""
    project = utils.get_project(project)
    c = Client(project=project)
    data = []
    headers = []
    filter_snippet = " (n=%s)" % n
    if len(filters) > 0:
        filter_snippet = " with filter %s" % str(filters) + filter_snippet
    for e in c.environments.list(n, filters=filters):
        (e_data, headers) = format_environment_data(e)
        data.append(e_data)
    typer.secho(
        "Found %s environments in project %s%s:" % (len(data), project, filter_snippet),
        fg="blue",
    )
    utils.print_table(data, headers, style=utils.get_style(style))


@app.command("get")
@utils.exit_on_error
def get(
    environment: str = typer.Argument(..., help="Environment ID."),
    project: str = typer.Option(None, help="Project ID."),
    json: bool = typer.Option(False, "--json", help="Print full JSON representation."),
):
    """Get environment details."""
    project, environment = utils.get_project_and_environment(project, environment)
    c = Client(project=utils.get_environment_name(project, environment))

    env = c.environments.get(environment)
    if json:
        console = Console()
        console.print(env.to_dict())
    else:
        data = format_environment_data(env, zipped=True)
        typer.secho("\nRetrieving environment %s: \n" % (environment), fg="blue")
        utils.print_info(data)


@app.command("create")
@utils.exit_on_error
def create(
    environment: str = typer.Argument(..., help="Environment ID."),
    project: str = typer.Option(None, help="Project ID."),
    source: str = typer.Option(
        None,
        "--from",
        help="Fully-qualified Model, Feature Set, Project, or Environment name from which to create this Environment.",
    ),
    database: str = typer.Option(
        None,
        "--db",
        help="Database to use for environment. Default is the same database as prod.",
    ),
    schema: str = typer.Option(
        None,
        help="Schema to use for environment. Default is to postfix the prod schema by this environment name.",
    ),
    scheduling_enabled: bool = typer.Option(
        False,
        "--scheduling-enabled",
        help="Enable cron schedules for this environment.",
    ),
):
    """Create environment."""
    project, environment = utils.get_project_and_environment(project, environment)
    c = Client(project=utils.get_environment_name(project, environment))
    proj = c.projects.get(project)
    fs_type = proj.data_store.type
    conn_info = {}
    if database:
        conn_info["database"] = database
    if schema:
        if fs_type == "big_query":
            conn_info["dataset"] = schema
        else:
            conn_info["db_schema"] = schema
    try:
        if schema is not None or database is not None:
            data_store = {"type": fs_type}
            data_store[fs_type] = conn_info
            env = c.environments.create(
                id=environment,
                source=source,
                data_store=data_store,
                scheduling_enabled=scheduling_enabled,
            )
        else:
            env = c.environments.create(
                id=environment, source=source, scheduling_enabled=scheduling_enabled
            )
        typer.secho("Successfully created environment %s." % env.name, fg="green")
    except Exception as e:
        typer.secho(
            "Failed to create environment %s in project %s: %s"
            % (environment, project, str(e)),
            fg="red",
        )


@app.command("delete")
@utils.exit_on_error
def delete(
    environment: str = typer.Argument(..., help="Environment ID."),
    project: str = typer.Option("", help="Project ID."),
    force: bool = typer.Option(
        False, "--force", help="Force deletion. Skips confirmation."
    ),
    delete_schema: bool = typer.Option(
        False,
        "--delete-schema",
        help="Delete the underlying data warehouse schema (in BigQuery, the schema corresponds to the 'dataset').",
    ),
):
    """Delete environment."""
    project, environment = utils.get_project_and_environment(project, environment)
    c = Client(project=utils.get_environment_name(project, environment))

    if not force:
        force = typer.confirm(
            "Are you sure you want to delete environment %s in project %s? This will delete all resources in the environment."
            % (environment, project)
        )
    if force:
        env = c.environments.get(environment)
        env.delete(delete_schema=delete_schema)
        if c.config.environment == environment and c.config.project == project:
            typer.secho(
                "Since (%s) was the active environment in the active project, the active environment was switched to production."
                % environment,
                fg="blue",
            )
            c.config.set_environment("prod")
        typer.secho(
            "Successfully deleted environment %s in project %s."
            % (environment, project),
            fg="green",
        )
    else:
        typer.secho("Environment deletion cancelled! ", fg="green")


@app.command("update")
@utils.exit_on_error
def update(
    environment: str = typer.Argument(..., help="environment ID."),
    project: str = typer.Option(None, help="Project ID."),
    scheduling_enabled: bool = typer.Option(
        False, "--scheduling-enabled", help="Is scheduling enabled in the environment?"
    ),
    database: str = typer.Option(
        None,
        "--db",
        help="Database to use for environment. Default is the same database as prod.",
    ),
    schema: str = typer.Option(
        None,
        help="Schema to use for environment. Default is to postfix the prod schema by this environment name.",
    ),
):
    """Update environment name, database, or schema."""
    project, environment = utils.get_project_and_environment(project, environment)
    c = Client(project=utils.get_environment_name(project, environment))

    env = c.environments.get(environment)
    try:
        if schema is not None or database is not None:
            fs_type = env.data_store.type
            conn_info = {}
            if database:
                conn_info["database"] = database
            if schema:
                if fs_type == "big_query":
                    conn_info["dataset"] = schema
                else:
                    conn_info["db_schema"] = schema
            data_store = {"type": fs_type}
            data_store[fs_type] = conn_info
            env.update(scheduling_enabled=scheduling_enabled, data_store=data_store)
        else:
            env.update(scheduling_enabled=scheduling_enabled)
        typer.secho(
            "Successfully updated environment %s in project %s."
            % (environment, project),
            fg="green",
        )
    except Exception as e:
        typer.secho(
            "Failed to update environment %s in project %s: %s"
            % (environment, project, str(e)),
            fg="red",
        )


@app.command("seed")
@utils.exit_on_error
def seed(
    paths: List[Path] = typer.Argument(
        ..., exists=True, dir_okay=True, help="Path to CSV file(s)."
    ),
    environment: str = typer.Option(None, help="environment ID."),
    project: str = typer.Option(None, help="Project ID."),
    table: str = typer.Option(None, help="Table name to create and store data into. "),
):
    """Upload data to data warehouse.

    Uploads all given CSV files to the environment and create tables loaded with data.
    The PATHS can be a list of CSV files to be uploaded.
    """
    projects.seed(paths=paths, environment=environment, project_id=project, table=table)


@app.command("export")
@utils.exit_on_error
def export(
    environment: str = typer.Argument(..., help="environment ID."),
    project_id: str = typer.Option(None, "--project", help="Project ID."),
    path: Path = typer.Option(
        None,
        help="Directory to save projects resources into. Defaults to ./<project_id> if not provided.",
    ),
    # include_predictions: bool = typer.Option(False, "--predictions", hidden=True, help="Include csv of latest predictions for each model"),
):
    """
    Exports project YAMLs to a directory. Akin to 'git clone' for continual.
    If you wish to export a single feature set or model, see
    `continual feature-sets export` or `continual model export` instead.
    """
    project_id, environment = utils.get_project_and_environment(project_id, environment)
    p_slug = utils.get_environment_name(project_id, environment)
    client = Client(project=p_slug)
    env = client.environments.get(p_slug)

    projects.export_project(path, "%s_%s" % (project_id, environment), env)



from __future__ import annotations
from typing import List, Optional
from continual.rpc.management.v1 import management_pb2
from continual.rpc.management.v1 import types
from continual.python.sdk.resource import Resource
from continual.python.sdk.manager import Manager
from continual.python.sdk.iterators import Pager
from continual.python.sdk.waiter import wait
from continual.python.sdk.events import EventManager


class PromotionManager(Manager):
    """Manages promotion resources."""

    name_pattern: str = "projects/{project}/models/{model}/promotions/{promotion}"

    def get(self, id: str) -> Promotion:
        """Get promotion.

        Arguments:
            id: Promotion name or id.

        Returns
            A promotion.
        """
        req = management_pb2.GetPromotionRequest(name=self.name(id))
        resp = self.client._management.GetPromotion(req)
        return Promotion.from_proto(resp, client=self.client)

    def list(
        self,
        page_size: Optional[int] = None,
        filters: List[str] = None,
        all_projects: bool = False,
    ) -> List[Promotion]:
        """List promotions.

        Arguments:
            page_size: Number of items to return.

        Returns:
            A list of promotions.
        """
        req = management_pb2.ListPromotionsRequest(
            parent=self.parent,
            page_size=page_size,
            filters=filters,
            all_projects=all_projects,
        )
        resp = self.client._management.ListPromotions(req)
        return [Promotion.from_proto(x, client=self.client) for x in resp.promotions]

    def list_all(self) -> Pager[Promotion]:
        """List all promotions.

        Pages through all promotions using an iterator.

        Returns:
            A iterator of all promotions.
        """

        def next_page(next_page_token):
            req = management_pb2.ListPromotionsRequest(
                parent=self.parent, page_token=next_page_token
            )
            resp = self.client._management.ListPromotions(req)
            return (
                [Promotion.from_proto(x, client=self.client) for x in resp.promotions],
                resp.next_page_token,
            )

        return Pager(next_page)


class Promotion(Resource, types.Promotion):
    """Promotion resource."""

    name_pattern: str = "projects/{project}/models/{model}/promotions/{promotion}"
    manager: PromotionManager

    events: EventManager
    """Event manager."""

    def _init(self):
        self.manager = PromotionManager(parent=self.parent, client=self.client)
        self.events = EventManager(parent=self.name, client=self.client)

    def wait(
        self,
        echo: bool = False,
        timeout: Optional[int] = None,
    ) -> Promotion:
        """Wait for promotion state transition to complete.

        Arguments:
            echo: Display progress.
            timeout: Timeout in second.

        Returns:
            An updated model version.
        """
        return wait(lambda: self.manager.get(self.id), echo=echo, timeout=timeout)



#!/usr/bin/env python3

import tarfile
import requests
import typer
import os
import yaml

from continual.python.cli import utils
from continual.python.sdk.featurestore_config import FeaturestoreConfig
from continual.python.sdk.client import Client
from continual.rpc.management.v1.types import FieldType

from rich.console import Console
from typing import List
from pathlib import Path
from enum import Enum

app = typer.Typer(help="Manage projects.")


class SchemaType(str, Enum):
    FeatureSet = "FeatureSet"
    Model = "Model"


def format_project_data(p, zipped=False):
    create_time = p.create_time.replace(microsecond=0)
    update_time = p.update_time.replace(microsecond=0)
    org = p.organization.split("/")[-1]
    fs_type = p.data_store.type
    c = Client()
    project_id = p.id
    current_project = c.config.project
    if current_project and current_project == p.name:
        project_id = f"{project_id} (active)"
    data = [
        project_id,
        p.display_name,
        org,
        fs_type,
        p.summary.feature_set_count,
        p.summary.feature_count,
        p.summary.model_count,
        p.summary.model_version_count,
        p.summary.prediction_count,
        create_time,
        # update_time,
    ]
    headers = [
        "ID",
        "Display Name",
        "Organization",
        "Data Warehouse",
        "Feature Sets",
        "Features",
        "Models",
        "Model Versions",
        "Predictions",
        "Created",
        # "Updated",
    ]

    if zipped:
        return tuple(
            [x[0], x[1]] for x in (zip(headers, data))
        )  # for some reason list(zip) causes issues, so ...
    else:
        return (data, headers)


# use callback to run list command if nothing is passed in
@app.callback(invoke_without_command=True)
def default(ctx: typer.Context):
    if ctx.invoked_subcommand is not None:
        return
    else:
        list(n=30, filters=[], style=None)


@app.command("list")
@utils.exit_on_error
def list(
    n: int = typer.Option(30, "--num", "-n", help="Number of records to show."),
    filters: List[str] = typer.Option([], "--filter", "-f", help="List of filters."),
    style: utils.ContinualStyle = typer.Option(None, help="Color to use for list."),
):
    """List projects."""
    c = Client()
    data = []
    headers = []
    filter_snippet = " (n=%s)" % n
    if len(filters) > 0:
        filter_snippet = " with filter %s" % str(filters) + filter_snippet
    for p in c.projects.list(n, filters=filters):
        (p_data, headers) = format_project_data(p)
        data.append(p_data)
    typer.secho("Found %s projects%s:" % (len(data), filter_snippet), fg="blue")
    utils.print_table(data, headers, style=utils.get_style(style))


@app.command("get")
@utils.exit_on_error
def get(
    project_id: str = typer.Argument(..., help="Project ID."),
    json: bool = typer.Option(False, "--json", help="Print full JSON representation."),
):
    """Get project details."""
    project_id = utils.get_project(project_id)  # handle passing in project@env
    c = Client(project=project_id)
    project = c.projects.get(project_id)
    if json:
        console = Console()
        console.print(project.to_dict())
    else:
        data = format_project_data(project, zipped=True)
        typer.secho("\nRetrieving project %s: \n" % (project_id), fg="blue")
        utils.print_info(data)


@app.command("create")
@utils.exit_on_error
def create(
    project: str = typer.Argument(..., help="Name of project."),
    org: str = typer.Option(..., help="Name of organization.", show_default=False),
    feature_store: str = typer.Option(
        ..., help="Feature store config name.", show_default=False
    ),
):
    """Create project."""
    try:
        c = Client()
        c.projects.get(project)
        typer.secho(
            "Project %s already exists. Please try another name.\n" % project, fg="red"
        )
    except:
        fs_config = FeaturestoreConfig()

        try:
            feature_store_dict = dict(fs_config._cp[feature_store])
        except:
            typer.secho(
                "Unable to look up feature store configuration for %s. Exiting..."
                % feature_store,
                fg="red",
            )
            raise typer.Exit(code=1)
        try:
            proj = utils.get_or_create_project(project, org, feature_store_dict)
            typer.secho(
                "Successfully created project %s with feature store %s."
                % (proj.id, feature_store),
                fg="green",
            )
        except Exception as e:
            typer.secho(
                "Error building project with feature store configuration %s: %s"
                % (feature_store, str(e)),
                fg="red",
            )
            raise typer.Exit(code=1)


@app.command("delete")
@utils.exit_on_error
def delete(
    project_id: str = typer.Argument(..., help="Project ID."),
    force: bool = typer.Option(
        False, "--force", help="Force deletion. Skips confirmation."
    ),
    delete_schema: bool = typer.Option(
        False,
        "--delete-schema",
        help="Delete the underlying data warehouse schema (in BigQuery, the schema corresponds to the 'dataset').",
    ),
):
    """Delete project."""
    project_id = utils.get_project(project_id)
    if not force:
        force = typer.confirm(
            "Are you sure you want to delete the project %s?" % project_id
        )
    if force:
        c = Client(project=project_id)
        project = c.projects.get(project_id)
        project.delete(delete_schema=delete_schema)
        typer.secho("Successfully deleted project %s." % project_id, fg="green")
    else:
        typer.secho("Skipping deletion!")


@app.command("update")
@utils.exit_on_error
def update(
    project_id: str = typer.Argument(..., help="Project ID."),
    new_name: str = typer.Option(None, help="New project name"),
    feature_store: str = typer.Option(
        None, help="Feature store config name.", show_default=False
    ),
):
    """Update project."""
    project_id = utils.get_project(project_id)
    c = Client(project=project_id)
    project = c.projects.get(project_id)
    feature_store_dict = project.data_store
    if feature_store:
        fs_config = FeaturestoreConfig()

        try:
            feature_store_dict = dict(fs_config._cp[feature_store])
        except:
            typer.secho(
                "Unable to look up feature store configuration for %s. Exiting..."
                % feature_store,
                fg="red",
            )
            raise typer.Exit(code=1)
    if new_name is None:
        new_name = project.display_name
    project.update(display_name=new_name, data_store=feature_store_dict)
    typer.secho("Successfully updated project %s." % project_id, fg="green")


@app.command("seed")
@utils.exit_on_error
def seed(
    paths: List[Path] = typer.Argument(
        ..., exists=True, dir_okay=True, help="Path to CSV file(s)."
    ),
    project_id: str = typer.Option(None, "--project", help="Project ID."),
    environment: str = typer.Option(None, "--env", hidden=True, help="Environment ID."),
    table: str = typer.Option(None, help="Table name to create and store data into. "),
):
    """Upload data to data warehouse.

    Uploads all given CSV files to the project and create tables loaded with data.
    The PATHS can be a list of CSV files to be uploaded.
    """

    project, environment = utils.get_project_and_environment(project_id, environment)
    client = Client(project=utils.get_environment_name(project, environment))
    env = client.environments.get(environment)

    if len(env.data_store.type) == 0:
        typer.secho(
            "No datastore set for %s. Please update via `continual environments update` or `continual projects update` or in the Environments menu in the UI."
            % env.id,
            fg="red",
        )
        raise typer.Exit(code=1)

    # typer path type is PosixPath, which our SDK doesn't expect. Only works w/ strings or list of strings.
    if len(paths) > 0:
        paths = [str(path) for path in paths]

    for path in paths:
        typer.secho("Seeding file %s" % path, fg="blue")
        try:
            (_, new_table) = env.seed(path, table_name=table)
            (schema, database) = utils.get_datastore_database_and_schema(env.data_store)
            if database:
                database = database + "."
            if schema:
                schema = schema + "."
            typer.secho(
                "Successfully seeded file %s to %s%s%s"
                % (
                    path,
                    database,
                    schema,
                    new_table,
                ),
                fg="green",
            )
        except Exception as e:
            typer.secho("Failed to seed file %s: %s" % (path, str(e)), fg="red")
            continue


@app.command("infer-schema")
@utils.exit_on_error
def infer_schema(
    schema_type: SchemaType = typer.Argument(..., help="Schema type to create."),
    project: str = typer.Option(None, help="Project ID."),
    index: str = typer.Option(
        ..., help="Column that corresponds to the model/feature set index."
    ),
    time_index: str = typer.Option(
        None, help="Column that corresponds to the model/feature set time index."
    ),
    split: str = typer.Option(
        None,
        help="Column that corresponds to the model user-defined split (model only)",
    ),
    target: str = typer.Option(None, help="Target column to predict (models only)."),
    entity: str = typer.Option(None, help="Entity to associate resource with."),
    table: str = typer.Option(None, help="Table name."),
    query: str = typer.Option(None, help="SQL query."),
    file: str = typer.Option(None, hidden=True, help="File name."),
    path: Path = typer.Option(None, help="Path to save schema."),
):
    """
    Infers schema from a table or SQL query.
    """
    project = utils.get_project(project)
    c = Client(project=project)
    schema = c.infer_schema(
        project=project, table=table, query=query, file=file, type=schema_type.value
    )

    ##we should probably generalize utils.format_yaml and then use it here as well.
    if table is not None:
        name = table.replace(".", "_")
    elif query is not None:
        name = "SQL_QUERY_INFERRED"
    else:
        name = file.split("/")[-1].split(".")[0]

    columns = []
    for col in schema:
        columns.append(
            {
                "name": col.name,
                "type": FieldType.from_proto(col.type).value,
                # "description" : col.description,
            }
        )

    new_schema = {
        "type": schema_type.value,
        "name": name,
        "description": "Inferred via Continual CLI.",
        "owners": [c.config.email],
        "index": index,
    }

    if time_index:
        new_schema["time_index"] = time_index

    if entity:
        new_schema["entity"] = entity

    if schema_type == SchemaType.Model:
        if target:
            new_schema["target"] = target
        else:
            typer.secho(
                "No target column found. You will need to declare a target before attemping a push.",
                fg="yellow",
            )

        if split:
            new_schema["split"] = split

    new_schema["columns"] = columns

    if table:
        new_schema["table"] = table
    elif query:
        new_schema["query"] = query
    else:
        typer.secho(
            "No table or query found. One of these is required before attemping a push.",
            fg="yellow",
        )

    yaml_text = yaml.dump(new_schema, sort_keys=False)

    if path is None:
        typer.secho("Successfully inferred schema for %s: \n" % table, fg="green")
        print(yaml_text)
    else:
        path = str(path)
        if (path[-4:] != ".yml") and (path[-5:] != ".yaml"):
            path = "%s/%s.yaml" % (path, name)
        with open(path, "w+") as f:
            f.write(yaml_text)
        typer.secho(
            "Successfully inferred schema for %s and saved to file %s." % (table, path),
            fg="green",
        )

        typer.secho(
            "You may can push this file into continual via 'continual push %s`." % path,
            fg="blue",
        )


@app.command("export")
@utils.exit_on_error
def export(
    project_id: str = typer.Option(None, "--project", help="Project ID."),
    path: Path = typer.Option(
        None,
        help="Directory to save projects resources into. Defaults to ./<project_id> if not provided.",
    ),
    # include_predictions: bool = typer.Option(False, "--predictions", hidden=True, help="Include csv of latest predictions for each model"),
):
    """
    Exports project YAMLs to a directory. Akin to 'git clone' for continual.
    If you wish to export a single feature set or model, see
    `continual feature-sets export` or `continual model export` instead.
    """
    project_id = utils.get_project(project_id)
    client = Client(project=project_id)
    project = client.projects.get(project_id)

    export_project(path, project_id, project)


def export_project(path, project_id, project):

    # use cwd/project_id if no path provided
    if path is None:
        path = os.path.join(os.getcwd(), project_id)
        if not os.path.exists(path):
            os.mkdir(path)

    try:
        feature_sets, models, extensions = project.export()
    except:
        typer.secho("Failed to export environment.", fg="red")
        exit(1)

    fs_dir = os.path.join(path, "feature_sets")
    model_dir = os.path.join(path, "models")
    extensions_dir = os.path.join(path, "extensions")
    # ßbpj_dir = os.path.join(path,"predictions")

    if len(feature_sets) > 0:
        if not os.path.exists(fs_dir):
            os.mkdir(fs_dir)
        fs_yamls = []
        for fs, schema in feature_sets.items():
            typer.secho(f"Saving feature set {fs}", fg="blue")
            fs_text = yaml.load(schema, Loader=yaml.FullLoader)
            fs_yaml = utils.format_yaml(fs_text)
            fs_yamls.append((fs, fs_yaml))
        utils.save_yamls(fs_yamls, fs_dir)
        typer.secho(
            "Successfully saved feature set yamls in directory: %s \n" % fs_dir,
            fg="green",
        )
    else:
        typer.secho("No feature sets found!", fg="blue")

    if len(models) > 0:
        if not os.path.exists(model_dir):
            os.mkdir(model_dir)
        model_yamls = []
        for model, schema in models.items():
            typer.secho(f"Saving model {model} ...", fg="blue")
            model_text = yaml.load(schema, Loader=yaml.FullLoader)
            model_yaml = utils.format_yaml(model_text)
            model_yamls.append((model, model_yaml))
        utils.save_yamls(model_yamls, model_dir)
        typer.secho(
            "Successfully saved model yamls in directory: %s \n" % model_dir,
            fg="green",
        )
    else:
        typer.secho("No models found!", fg="blue")

    if len(extensions) > 0:
        if not os.path.exists(extensions_dir):
            os.mkdir(extensions_dir)
        for extension, targz_url in extensions.items():
            res = requests.get(targz_url, stream=True)
            targz = tarfile.open(fileobj=res.raw, mode="r|gz")
            targz.extractall(extensions_dir)
            typer.secho(f"Saving extension {extension} to {extensions_dir}", fg="blue")
    else:
        typer.secho("No extensions found!", fg="blue")

    typer.secho("\nExport finished!", fg="green")

    # try:
    #     if include_predictions:
    #         if not os.path.exists(bpj_dir):
    #             os.mkdir(bpj_dir)
    #         for m in project.models.list_all():
    #             if m.latest_batch_prediction is not None:
    #                 typer.secho("Fetching latest batch prediction for model %s ... \n" %
    #                             m.name, fg="blue")
    #                 bpj = client.batch_prediction_jobs.get(m.latest_batch_prediction)
    #                 bpj.download(os.path.join(path,"%s_%s.csv" %(m.id, bpj.id)))
    #         typer.secho("Successfully saved latest predictions in directory %s: \n" %
    #                             bpj_dir, fg="green")
    # except Exception as e:
    #     typer.secho("Failed to get all latest predictions. Failed at prediction %s:%s \n" %
    #             (bpj.name, str(e)), fg="red")



import unittest
import mock

from continual.python.sdk.changes import Change

from continual.python.cli import utils


class AStringHaving(str):
    def __eq__(self, other):
        return self in other


class print_change(unittest.TestCase):
    def test_should_handle_non_default_environments(self):
        with mock.patch("typer.secho") as sut:
            utils.print_change(
                Change(), "some_project", "some_non_default_env", "message"
            )
            print(sut.call_args_list)
            sut.assert_any_call(
                "Updating some_project@some_non_default_env:", fg=mock.ANY
            )
            sut.assert_any_call(
                "Operations for  in project 'some_project', environment 'some_non_default_env':",
                fg=mock.ANY,
            )
            sut.assert_any_call(
                AStringHaving("/projects/some_project@some_non_default_env/changes/"),
                fg=mock.ANY,
            )

    def test_should_handle_default_environment(self):
        with mock.patch("typer.secho") as sut:
            utils.print_change(Change(), "some_project", "prod", "message")
            print(sut.call_args_list)
            sut.assert_any_call("Updating some_project@production:", fg=mock.ANY)
            sut.assert_any_call(
                "Operations for  in project 'some_project', environment 'production':",
                fg=mock.ANY,
            )
            sut.assert_any_call(
                AStringHaving("/projects/some_project/changes/"),
                fg=mock.ANY,
            )

    def test_should_handle_default_environment_name(self):
        with mock.patch("typer.secho") as sut:
            utils.print_change(
                Change(), "some_project", "projects/some_project", "message"
            )
            print(sut.call_args_list)
            sut.assert_any_call("Updating some_project@production:", fg=mock.ANY)
            sut.assert_any_call(
                "Operations for  in project 'some_project', environment 'production':",
                fg=mock.ANY,
            )
            sut.assert_any_call(
                AStringHaving("/projects/some_project/changes/"),
                fg=mock.ANY,
            )



import os
import logging
import toml

CONFIG_ENV = "CONFIG_ENV"
CONFIG_LOCAL_BASE = "CONFIG_LOCAL_BASE"


def get_config(config_type):

    env = os.environ.get(CONFIG_ENV, "dev")
    if env != "dev" and env != "prod" and env != "stage" and env != "local":
        # TODO Exception.
        logging.info("In config - Exiting -  " + env + "-")
        return None

    overrride_file = ""
    config_base_file = "/app/{0}-config-base.toml".format(config_type)
    config_file = "/app/{0}-config.toml".format(config_type)
    local_base = "."

    if env == "local":
        # get local base
        if os.environ.get(CONFIG_LOCAL_BASE, None) is not None:
            local_base = os.environ.get(CONFIG_LOCAL_BASE, None)

    if env == "dev":
        overrride_file = "/app/{0}-config-override.toml".format(config_type)
    elif env == "local":
        config_base_file = "{0}/deployment/config/base/{1}.toml".format(
            local_base, config_type
        )
        config_file = "{0}/deployment/config/dev/{1}.toml".format(
            local_base, config_type
        )
        overrride_file = "{0}/deployment/config/dev/{1}-override.toml".format(
            local_base, config_type
        )

    cfg = {}
    with open(config_base_file, "r") as inp:
        c = toml.load(inp)
        cfg.update(c)

    with open(config_file, "r") as inp:
        c = toml.load(inp)
        cfg.update(c)

    if os.path.isfile(overrride_file):
        with open(overrride_file, "r") as inp:
            c = toml.load(inp)
            cfg.update(c)

    return cfg



#!/usr/bin/env python3

import typer
import json

from continual.python.sdk.featurestore_config import FeaturestoreConfig
from continual.python.cli import utils

from enum import Enum


app = typer.Typer(help="Manage local data warehouse configurations.")

allowed_fs_types = ["snowflake", "continual"]


class FeaturestoreType(str, Enum):
    snowflake = "snowflake"
    # continual = "continual"
    bigquery = "bigquery"
    redshift = "redshift"
    # synapse = "synapse"


@app.command("list")
@utils.exit_on_error
def list(
    details: bool = typer.Option(False, help="Print full details."),
    unmask: bool = typer.Option(
        False, help="Print passwords in plaintext", hidden=True
    ),
):
    """List all data warehosue configs."""
    fsconfig = FeaturestoreConfig()
    fsconfig.list(details, unmask)


@app.command("print")
@utils.exit_on_error
def print(
    name: str = typer.Argument(..., help="Data warehouse config name to retrieve.")
):
    """Print a single data warehouse config."""
    fsconfig = FeaturestoreConfig()
    fsconfig.get(name)


@app.command("create")
@utils.exit_on_error
def create(
    name: str = typer.Argument(..., help="Data warehouse config name to create."),
    type: FeaturestoreType = typer.Option(None, help="Data warehouse type."),
    dict_string: str = typer.Option(
        ..., help="String of dictionary values to use for data warehouse configuration."
    ),
):
    """
    Creates a data warehouse configuration with the dictionary string passed as parameters.
    The parameters needed are dependent upon your data warehouse type.

    Snowflake: DATABASE, USERNAME, PASSWORD, HOST, WAREHOUSE, DB_SCHEMA, ROLE

    BigQuery: AUTH_FILE, DATASET

    REDSHIFT: DATABASE, USERNAME, PASSWORD, HOST, PORT, DB_SCHEMA

    Example:

            continual warehouse-config myFS --type snowflake --dict-string '{"DATABASE" : "myDB", "USERNAME" : "myUser", "PASSWORD" : "myPass", "HOST" : "myHost", "WAREHOUSE" : "myWarehouse"}'

    Currently only types snowflake, bigquery, and redshift are supported.
    """
    fsconfig = FeaturestoreConfig()
    fsconfig.delete(name)
    if type.lower() in allowed_fs_types:
        config = json.loads(dict_string)
        config = {x.upper(): config[x] for x in config}
        config["TYPE"] = type.value.lower()
        fsconfig.create(name, config)
        typer.secho(
            "Successfully created data warehouse configuration %s" % name, fg="green"
        )
    else:
        typer.secho("Featurestore type %s not supported." % type.value, fg="red")


@app.command("delete")
@utils.exit_on_error
def delete(
    name: str = typer.Argument(..., help="Data warehouse config name to delete.")
):
    """Deletes a single data warehouse config."""
    fsconfig = FeaturestoreConfig()
    fsconfig.delete(name)
    typer.secho("Successfully deleted section %s" % name, fg="green")


@app.command("update")
@utils.exit_on_error
def update(
    name: str = typer.Argument(..., help="Data warehouse config name to retrieve."),
    key: str = typer.Option(..., help="Data warehouse config key to set."),
    value: str = typer.Option(..., help="Data warehouse config value to use."),
):
    """Updates a single key-value pair in a data warehouse config."""

    fsconfig = FeaturestoreConfig()
    key = key.upper()
    fsconfig.update(name, key, value)
    typer.secho("Successfully updated section %s" % name, fg="green")



from __future__ import annotations
from typing import List, Optional
from continual.rpc.management.v1 import management_pb2
from continual.rpc.management.v1 import types
from continual.python.sdk.resource import Resource
from continual.python.sdk.manager import Manager
from continual.python.sdk.iterators import Pager
from continual.python.sdk.events import EventManager


class ExperimentManager(Manager):
    """Manages experiment resources."""

    name_pattern: str = (
        "projects/{project}/models/{model}/versions/{version}/experiments/{experiment}"
    )

    def get(self, id: str) -> Experiment:
        """Get experiment.

        Arguments:
            id: Experiment name or id.

        Returns
            An experiment.
        """
        req = management_pb2.GetExperimentRequest(name=self.name(id))
        resp = self.client._management.GetExperiment(req)
        return Experiment.from_proto(resp, client=self.client)

    def list(
        self,
        page_size: Optional[int] = None,
        filters: List[str] = None,
        all_projects=False,
    ) -> List[Experiment]:
        """List experiments.

        Arguments:
            page_size: Number of items to return.

            filters: List of filters to apply to experiment. Can be:
            - state  (i.e. state:FAILED)


        Returns:
            A list of experiments.
        """
        req = management_pb2.ListExperimentsRequest(
            parent=self.parent,
            page_size=page_size,
            filters=filters,
            all_projects=all_projects,
        )
        resp = self.client._management.ListExperiments(req)
        return [Experiment.from_proto(x, client=self.client) for x in resp.experiments]

    def list_all(self) -> Pager[Experiment]:
        """List all experiments.

        Pages through all experiments using an iterator.

        Returns:
            A iterator of all experiments.
        """

        def next_page(next_page_token):
            req = management_pb2.ListExperimentsRequest(
                parent=self.parent, page_token=next_page_token
            )
            resp = self.client._management.ListExperiments(req)
            return (
                [
                    Experiment.from_proto(x, client=self.client)
                    for x in resp.experiments
                ],
                resp.next_page_token,
            )

        return Pager(next_page)


class Experiment(Resource, types.Experiment):
    """Experiment resource."""

    manager: ExperimentManager
    events: EventManager
    """Event manager."""

    name_pattern: str = (
        "projects/{project}/models/{model}/versions/{version}/experiments/{experiment}"
    )

    def _init(self):
        self.manager = ExperimentManager(parent=self.parent, client=self.client)
        self.events = EventManager(parent=self.name, client=self.client)

    def cancel(self) -> None:
        """Cancel experiment."""
        raise NotImplementedError("Not yet implemented.")

    def wait(self) -> None:
        """Wait for experiment to complete."""
        raise NotImplementedError("Not yet implemented.")



import json
import unittest

import mock
import requests
from six.moves import http_client


def _make_credentials():
    import google.auth.credentials

    return mock.Mock(spec=google.auth.credentials.Credentials)


def _make_connection(*responses):
    import google.cloud.storage._http
    from google.cloud.exceptions import NotFound

    mock_conn = mock.create_autospec(google.cloud.storage._http.Connection)
    mock_conn.user_agent = "testing 1.2.3"
    mock_conn.api_request.side_effect = list(responses) + [NotFound("miss")]
    return mock_conn


def _make_response(status=http_client.OK, content=b"", headers={}):
    response = requests.Response()
    response.status_code = status
    response._content = content
    response.headers = headers
    response.request = requests.Request()
    return response


def _make_json_response(data, status=http_client.OK, headers=None):
    headers = headers or {}
    headers["Content-Type"] = "application/json"
    return _make_response(
        status=status, content=json.dumps(data).encode("utf-8"), headers=headers
    )


def _make_requests_session(responses):
    session = mock.create_autospec(requests.Session, instance=True)
    session.request.side_effect = responses
    return session


class TestGCSClient(unittest.TestCase):
    @staticmethod
    def _get_target_class():
        from continual.python.blob.gcs import GoogleCloudStorage

        return GoogleCloudStorage

    def _make_one(self, *args, **kw):
        return self._get_target_class()(*args, **kw)

    def test_valid_resource_url(self):
        EMPTY_URL = ""
        self.assertFalse(self._get_target_class().is_valid_url(EMPTY_URL))

        with self.assertRaises(ValueError):
            client = self._make_one(EMPTY_URL)

        NO_BUCKET_URL = "gs://"
        self.assertFalse(self._get_target_class().is_valid_url(NO_BUCKET_URL))

        with self.assertRaises(ValueError):
            client = self._make_one(NO_BUCKET_URL)

        VALID_URL = "gs://bucket"
        self.assertTrue(self._get_target_class().is_valid_url(VALID_URL))

        client = self._make_one(VALID_URL)
        self.assertEqual(client.bucket_name, "bucket")

    def test_ctor_connection_type(self):
        from google.cloud._http import ClientInfo
        from google.cloud.storage._http import Connection

        PROJECT = "continual-fake"
        RESOURCE_URL = "gs://continual-test"
        credentials = _make_credentials()

        client = self._make_one(RESOURCE_URL, project=PROJECT, credentials=credentials)

        client = client.client

        self.assertEqual(client.project, PROJECT)
        self.assertIsInstance(client._connection, Connection)
        self.assertIs(client._connection.credentials, credentials)
        self.assertIsNone(client.current_batch)
        self.assertEqual(list(client._batch_stack), [])
        self.assertIsInstance(client._connection._client_info, ClientInfo)
        self.assertEqual(
            client._connection.API_BASE_URL, Connection.DEFAULT_API_ENDPOINT
        )


if __name__ == "__main__":
    # TODO(sahil): add testing for upload and download
    unittest.main()



import typer

from continual.python.cli import utils
from continual.python.sdk.client import Client

from typing import List
from rich.console import Console
from continual.services.compute_services.pipelines.batchprediction.batch_prediction_runner import (
    BatchPredictionRunner,
)

from continual.services.compute_services.pipelines.training.training_runner import (
    TrainingPipelineRunner,
)

app = typer.Typer(help="Provides tools and processes for extension development.")


@app.command("run")
@utils.exit_on_error
def run(
    project: str = typer.Option(None, help="Project ID."),
    environment: str = typer.Option(None, "--env", help="Environment ID."),
    model: str = typer.Argument(..., help="ID for the model to run locally"),
    paths: List[str] = typer.Option(
        None,
        "--paths",
        exists=True,
        dir_okay=True,
        help="Paths to YAML files or directories.",
    ),
    train: bool = typer.Option(False, "--train", help="Runs training if set."),
    trained_model_version: str = typer.Option(
        None,
        "--model-version",
        help="Trained model version name if only running batchprediction",
    ),
    batchpredict: bool = typer.Option(
        False, "--batchpredict", help="Runs batchprediction if set."
    ),
    to_json: bool = typer.Option(
        False, "--json", help="Show full JSON representation."
    ),
):
    """Run local training and batchprediction.

    Command will execute the following steps:
    * Push all YAMLS in `path`, creating any models, featuresets and extensions
    * If --train is specified, run training using the model pipeline specified in the YAML for the model `model_id`.
    * If --batchpredict is specified, run prediction using the model pipeline specified in the YAML for the model `model_id`.
    """

    project, environment = utils.get_project_and_environment(project, environment)
    client = Client(project=utils.get_environment_name(project, environment))
    model_version_name = None

    # Step 1 push without training
    if paths:
        changes = client.changes.push(paths=paths, train_none=True)
        changes.wait(timeout=3600)

    # Step 2, if train is specified create a local model version
    if train:
        try:
            local_model_version = client.models.create_local_dev_model_version(id=model)
            model_version_name = local_model_version.name

            client.set_logging_resource(resource=model_version_name)
            console = Console()
            if to_json:
                console.print(local_model_version.to_dict())
            else:
                typer.secho(
                    "Created model version %s \n" % (model_version_name),
                    fg="blue",
                )

            TrainingPipelineRunner(
                client=client,
                model_version_name=model_version_name,
            ).run()
        except Exception as e:
            typer.secho(
                f"Training failed with exception: {str(e)} \n",
                fg="red",
            )
    else:
        model_version_name = trained_model_version

    if batchpredict:
        if not model_version_name:
            typer.secho(
                "Error: Local batchprediction requires --train to be set OR --model-version to specified.",
                fg="red",
            )
            return
        try:
            local_batch_prediction = client.models.create_local_dev_batch_prediction(
                id=model, model_version_name=model_version_name
            )
            client.set_logging_resource(resource=local_batch_prediction.name)

            console = Console()
            if to_json:
                console.print(local_batch_prediction.to_dict())
            else:
                typer.secho(
                    "Created batch prediction %s \n" % (local_batch_prediction.name),
                    fg="blue",
                )
            BatchPredictionRunner(
                client=client,
                batch_prediction_name=local_batch_prediction.name,
            ).run()
        except Exception as e:
            typer.secho(
                f"Batchprediction failed with exception: {str(e)} \n",
                fg="red",
            )



from __future__ import annotations
from typing import List, Optional, Tuple
from git import Repo

import pandas as pd
import json

from continual.rpc.management.v1 import management_pb2
from continual.rpc.management.v1 import types
from continual.python.sdk.resource import Resource
from continual.python.sdk.manager import Manager
from continual.python.sdk.iterators import Pager
from continual.python.sdk.ingestions import Ingestion, IngestionManager
from continual.python.sdk.events import EventManager


class FeatureSetManager(Manager):
    """Manages feature set resources."""

    name_pattern: str = "projects/{project}/featureSets/{feature_set}"

    def get(self, id: str, parent: Optional[str] = None) -> FeatureSet:
        """Get feature set.

        Use @version in id to get a specific feature set id.

        Arguments:
            id: Feature set name or id.

        Returns
            A feature set.
        """
        req = management_pb2.GetFeatureSetRequest(name=self.name(id, parent=parent))
        resp = self.client._management.GetFeatureSet(req)
        return FeatureSet.from_proto(resp, client=self.client)

    def list(
        self,
        page_size: Optional[int] = None,
        filters: List[str] = None,
        all_projects=False,
    ) -> List[FeatureSet]:
        """List feature sets.

        Arguments:
            page_size: Number of items to return.

        Returns:
            A list of feature sets.
        """
        req = management_pb2.ListFeatureSetsRequest(
            parent=self.parent,
            page_size=page_size,
            filters=filters,
            all_projects=all_projects,
        )
        resp = self.client._management.ListFeatureSets(req)
        return [FeatureSet.from_proto(x, client=self.client) for x in resp.feature_sets]

    def list_all(self) -> Pager[FeatureSet]:
        """List all feature sets.

        Pages through all feature sets using an iterator.

        Returns:
            A iterator of all feature sets.
        """

        def next_page(next_page_token):
            req = management_pb2.ListFeatureSetsRequest(
                parent=self.parent, page_token=next_page_token
            )
            resp = self.client._management.ListFeatureSets(req)
            return (
                [
                    FeatureSet.from_proto(x, client=self.client)
                    for x in resp.feature_sets
                ],
                resp.next_page_token,
            )

        return Pager(next_page)

    def delete(self, id: str) -> None:
        """Delete an feature set.

        Arguments:
            id: Feature set name or id.
        """

        req = management_pb2.DeleteFeatureSetRequest(name=self.name(id))
        self.client._management.DeleteFeatureSet(req)

    def get_data(
        self, feature_set: FeatureSet, page_size: int = None, page_token: str = None
    ) -> Tuple[pd.DataFrame, str]:
        """Get data for the featureset

        Arguments:
            page_size: number of rows to fetch.
            page_token: starting token for the next page of data.

        Returns:
            A dictionary of features.
        """
        req = management_pb2.GetDataRequest(
            name=feature_set.name,
            page_size=page_size,
            page_token=page_token,
        )

        resp = self.client._management.GetData(req)
        df = None
        rows = []
        if resp.value is not None:
            for k in resp.value:
                feature_dict = json.loads(k)
                rows.append(feature_dict)

        if len(rows) > 0:
            df = pd.DataFrame(rows)

        return df, resp.next_page_token

    def diff(
        self,
        right_yaml: str,
        right_project: str,
        right_environment: str,
        right_resource: str,
        left_yaml: str,
        left_project: str,
        left_environment: str,
        left_resource: str,
        models: bool = False,
    ):
        """Get data for the featureset

        Arguments:
            right_yaml
            right_project
            right_environment
            right_feature_set
            left_yaml
            left_project
            left_environment
            left_featureset

        Returns:
            Diff Report
        """
        right = None
        left = None
        default_project = self.client.config.project.split("/")[-1]
        default_environment = self.client.config.environment
        if default_environment is None:
            repo = Repo(search_parent_directories=True)
            default_environment = repo.active_branch.name
            if default_environment is None:
                default_environment = "prod"

        if right_yaml is None:
            right = right_resource
            if right_project is None:
                right_project = default_project
            if right_environment is None:
                right_environment = default_environment
            right = "%s in project %s, environment %s" % (
                right,
                right_project,
                right_environment,
            )
            env = self.client.projects.get(right_project).environments.get(
                right_environment
            )
            if models:
                right_text = env.models.get(right_resource).schema_text
            else:
                right_text = env.feature_sets.get(right_resource).schema_text
        else:
            right = right_yaml
            with open(right_yaml, "r") as f:
                right_text = f.read()
        if left_yaml is None:
            left = left_resource
            if left_project is None:
                left_project = default_project
            if left_environment is None:
                left_environment = default_environment
            left = "%s in project %s, environment %s" % (
                left,
                left_project,
                left_environment,
            )
            env = self.client.projects.get(left_project).environments.get(
                left_environment
            )
            if models:
                left_text = env.models.get(left_resource).schema_text
            else:
                left_text = env.feature_sets.get(left_resource).schema_text
        else:
            left = left_yaml
            with open(left_yaml, "r") as f:
                left_text = f.read()

        req = management_pb2.YamlDiffRequest(
            right=right,
            left=left,
            right_text=right_text,
            left_text=left_text,
            parent=self.parent,
        )

        resp = self.client._management.GetYamlDiff(req)

        return (right, left, resp.text)


class FeatureSet(Resource, types.FeatureSet):
    """Feature set resource."""

    name_pattern: str = "projects/{project}/featureSets/{feature_set}"
    manager: FeatureSetManager

    ingestions: IngestionManager
    """Ingestion manager."""

    events: EventManager
    """Event manager."""

    def _init(self):
        self.manager = FeatureSetManager(parent=self.parent, client=self.client)
        self.ingestions = IngestionManager(parent=self.name, client=self.client)
        self.events = EventManager(parent=self.name, client=self.client)

    def delete(self) -> None:
        """Delete feature set."""
        self.manager.delete(self.name)

    def ingest(self) -> Ingestion:
        """Ingest feature set"""
        return self.manager.ingest(self.name)

    def list_versions(self, page_size: Optional[int] = None) -> List[FeatureSet]:
        """List feature set versions

        Arguments:
            page_size: number of elements

        Returns:
            List of feature set versions.
        """
        return self.manager.list_versions(id=self.name, page_size=page_size)

    def list_all_versions(self) -> Pager[FeatureSet]:
        """List all feature set versions

        Returns:
            An interator of feature set versions.
        """
        return self.manager.list_all_versions(id=self.name)

    # def training_data(self, model: str):
    #    return self.manager.training_data(self, self.id, model)

    def get_data(self, page_size=None, page_token=None):
        return self.manager.get_data(self, page_size, page_token)


# TODO: Add these methods back once data access is more clear.
# @authenticated
# @normalize_exceptions
# def get_all_entities(
#     self, name=None, visible_time=None, system_time=None, normalized=False,
# ):
#     """Fetches all entities for the given entity name from featurestore.

#     The version of entities returned is based on the visible_time and system_time.

#     If visible_time and system_time are not provided then latest instances of all
#      entities are returned

#     Args:
#         name: name of the entity to fetch.
#         visibile_time: default - now. Event time of of the entity. This allows for
#          going back to previous versions of entity.
#                For eg. if visible time is a date 3 months back from today, then
#                the data fetched will be of entity state known as of that date.
#         system_time: default - now. System time limits the entity's version to the
#          state of entity in continual at that time.
#         normalized: default - False. If set, then the data from related entities
#          are not returned.

#     Returns:
#         dataframe
#         A dataframe containing all entites in featurestore that match the given
#          time constraint.

#     Raises:

#     """

#     # Folling should be either configurable or server default.
#     page_size = 5000

#     entities = []
#     done = False
#     next_token = None
#     while not done:
#         req = featurestore_pb2.GetBatchRequest(
#             name=name,
#             project=self.name,
#             page_size=page_size,
#             page_token=next_token,
#             system_time_ms=system_time,
#             visible_time_ms=visible_time,
#             normalized=normalized,
#         )
#         batch_data = self._client._featurestore.GetBatch(req)
#         for val in batch_data.value:
#             e_dict = {}
#             for k in val.features:
#                 e_dict[k] = getattr(
#                     val.features[k], val.features[k].WhichOneof("kind")
#                 )
#             entities.append(e_dict)
#             next_token = batch_data.next_page_token
#             if next_token == "":
#                 done = True

#     df = pd.DataFrame.from_dict(entities)
#     return df

# @authenticated
# @normalize_exceptions
# def delete_entities(self, entity=None, name=None):
#     if entity is not None:
#         name = self.name + "/featureSets/" + entity
#     req = featurestore_pb2.DeleteEntitiesRequest(name=name)
#     self._client._featurestore.DeleteEntities(req)
#     return True



import typer

from continual.python.cli import utils
from continual.python.sdk.client import Client

from rich.console import Console
from datetime import datetime
from typing import List

app = typer.Typer(help="Manage Ingestions")


def format_ingestion_data(i, zipped=False, all_projects=False):
    feature_set = i.parent.split("/")[-1]
    end_time = i.finish_time
    start_time = i.start_time.replace(microsecond=0)
    if end_time:
        end_time = end_time.replace(microsecond=0)
        duration = (end_time - start_time).seconds
    else:
        end_time = "N/A"
        if i.state.value == "RUNNING":
            duration = (datetime.utcnow() - start_time).seconds
        else:
            duration = "N/A"
    if zipped:
        data = [
            i.id,
            i.name,
            i.state.value,
            feature_set,
            start_time,
            end_time,
            duration,
            i.ingested_row_count,
        ]
        headers = [
            "ID",
            "Name",
            "State",
            "Feature Set",
            "Started",
            "Finished",
            "Duration",
            "Ingested Rows",
        ]
        return tuple(
            [x[0], x[1]] for x in (zip(headers, data))
        )  # for some reason list(zip) causes issues, so ...
    else:
        data = [
            i.id,
            i.state.value,
            feature_set,
            start_time,
            end_time,
            duration,
            i.ingested_row_count,
        ]
        headers = [
            "ID",
            "State",
            "Feature Set",
            "Started",
            "Finished",
            "Duration (s)",
            "Ingested Rows",
        ]
        if all_projects:
            data.insert(0, i.parent.split("/")[1])
            headers.insert(0, "Project")
        return (data, headers)


@app.command("list")
@utils.exit_on_error
def list(
    feature_set: str = typer.Option(None, help="Feature Set ID"),
    project: str = typer.Option(None, help="Project ID"),
    n: int = typer.Option(30, "--num", "-n", help="Number of records to show"),
    filters: List[str] = typer.Option([], "--filter", "-f", help="List of filters"),
    all_projects: bool = typer.Option(
        False, "--all", "-a", help="Do not filter results by current project"
    ),
    style: utils.ContinualStyle = typer.Option(None, help="Color to use for report"),
):
    """List Ingestions

    filter can be one of the following:
    - state  (i.e. state:FAILED)
    """
    c = Client(project=project)
    if project is None:
        project = utils.get_default_project()
    fs_snippet = ""
    filter_snippet = ""
    project_snippet = "project %s" % project
    if feature_set is not None:
        c = c.feature_sets.get(feature_set)
        fs_snippet = "for feature set %s " % feature_set
    if len(filters) > 0:
        filter_snippet = " with filter %s" % str(filters)
    if all_projects:
        project_snippet = "all accessible projects"
    data = []
    headers = []
    for i in c.ingestions.list(n, filters=filters, all_projects=all_projects):
        (i_data, headers) = format_ingestion_data(i, all_projects=all_projects)
        data.append(i_data)
    typer.secho(
        "\nFound %s Ingestions %sin %s%s: "
        % (len(data), fs_snippet, project_snippet, filter_snippet),
        fg="blue",
    )
    if style is None:
        style = c.config.style
    style = utils.get_style(style)
    utils.print_table(data, headers, style=style)


@app.command("get")
@utils.exit_on_error
def get(
    ingestion: str = typer.Argument(..., help="Ingestion ID"),
    project: str = typer.Option(None, help="Project ID"),
    json: bool = typer.Option(False, "--json", help="Show full json representation?"),
):
    """Get Ingestion information"""
    c = Client(project=project)
    if project is None:
        project = utils.get_default_project()
    i = c.ingestions.get(ingestion)
    if json:
        console = Console()
        console.print(i.to_dict())
    else:
        data = format_ingestion_data(i, zipped=True)
        typer.secho("\nRetrieving ingestion %s: \n" % (ingestion), fg="blue")
        utils.print_info(data)


@app.command("cancel")
@utils.exit_on_error
def cancel(
    ingestion: str = typer.Argument(..., help="Ingestion ID"),
    project: str = typer.Option(None, help="Project ID"),
    feature_set: str = typer.Option(None, help="Feature set ID"),
):
    """Cancel an ingestion that is currently running"""
    c = Client(project=project)
    if project is None:
        project = utils.get_default_project()
    if "/" in ingestion:
        i_slug = ingestion
    else:
        i_slug = "projects/%s/featureSets/%s/ingetions/%s" % (
            project,
            feature_set,
            ingestion,
        )
    i = c.ingestions.get(i_slug)
    i.cancel()
    typer.secho("Successfully cancelled ingestion %s" % (i.name), fg="green")



from jinja2 import Template
from continual.python.sdk.query import Query
from continual.rpc.management.v1 import management_types_pb2
from google.protobuf.json_format import MessageToDict
import logging

time_index_template = Template(
    """(select {% if params.time_index %} max({{ params.time_index }}) from {{ params.table_name }} as {{ params.table_alias }} 
                      where {{ params.table_alias }}.{{ params.index }} = {{ params.model_alias }}.{{ params.match_index}}
                          {% if params.match_time_index %} and {{ params.table_alias }}.{{params.time_index}} <= {{ params.model_alias }}.{{params.match_time_index}} {% endif %} {% else %} {{params.match_time_index}} {% endif %}
                      ) as {{ params.qualified_time_index }} """
)

databricks_time_index_max_template = Template(
    """{% if params.time_index %} max({{ params.table_alias }}.{{params.match_time_index}}) as {{ params.qualified_time_index }} {% endif %}
    """
)
databricks_time_index_template = Template(
    """{% if params.time_index %} left join {{ params.table_name }} as {{ params.table_alias }} 
                on {{ params.table_alias }}.{{ params.match_index }} = {{ params.model_alias }}.{{ params.index }}
                and {{ params.table_alias }}.{{ params.match_time_index }} <= {{ params.model_alias }}.{{ params.time_index }}  {% endif %}
            """
)

train_template = Template(
    """
    {{ params.index_query }} SELECT time_index.* {% for fname in params.select_list %},{{ fname }} as {{params.select_list[fname]}} {% endfor %}
    FROM  time_index {% for fname in params.join_list %} LEFT JOIN {{fname}} AS {{params.join_list[fname][0]}} ON {{params.join_list[fname][1]}} {% endfor %}
    {% if params.filter_list %} WHERE {% for filter in params.filter_list %} {% if filter != params.filter_list[loop.index0] %} AND {% endif %} {{ filter}} {% endfor %} {% endif %}
    """
)

non_time_train_template = Template(
    """
    SELECT {% for fname in params.select_list %} {% if loop.index > 1 %},{% endif %} {{ fname }} as {{params.select_list[fname]}} {% endfor %}
    FROM   {% for fname in params.join_list %} {% if loop.index > 1 %} LEFT JOIN {% endif %}  {{fname}} AS {{params.join_list[fname][0]}} {% endfor %}
    {% if params.filter_list %} WHERE {% for filter in params.filter_list %} {% if filter != params.filter_list[loop.index0] %} AND {% endif %} {{ filter}} {% endfor %} {% endif %}
    """
)


def get_index(qdict):
    for field in qdict["fields"]:
        if ("type" in field and field["type"] == "INDEX") or (
            "index" in field and field["index"]
        ):
            return field["name"], field["qualified_name"]
    return None, None


def get_time_index_field(qdict):
    for field in qdict["fields"]:
        if ("type" in field and field["type"] == "TIME_INDEX") or (
            "time_index" in field and field["time_index"]
        ):
            return field["name"], field["qualified_name"]
    return None, None


def render_index_query(
    qdict,
    match_index,
    match_time_index,
    feature_store,
    table_alias="t1",
    column_alias_map=None,
    model_name=None,
    model_alias="t",
):
    project_id, table_name = feature_store.get_data_table_name(qdict["name"])
    full_name = qdict.get("table_name", None)
    if full_name is None:
        full_name = feature_store.db_schema + "." + table_name

    rel_index, rel_qualified_index = get_index(qdict)
    rel_time_index, rel_qualified_time_index = get_time_index_field(qdict)
    if rel_qualified_time_index is not None:
        ts_alias_name = rel_qualified_time_index.replace(":", "_")
        ts_alias_name = ts_alias_name.replace("$", "_")

    else:
        ts_alias_name = rel_qualified_time_index
    column_alias_map[ts_alias_name] = rel_qualified_time_index

    params = {
        "db_schema": feature_store.db_schema,
        "table_name": full_name,
        "index": rel_index,
        "time_index": rel_time_index,
        "match_index": match_index,
        "qualified_time_index": ts_alias_name,
        "match_time_index": match_time_index,
        "table_alias": table_alias,
        "model_alias": model_alias,
    }

    if feature_store.get_type() == "databricks":
        params["model_name"] = model_name or ""
        return (
            databricks_time_index_template.render(params=params),
            databricks_time_index_max_template.render(params=params),
        )
    else:
        return time_index_template.render(params=params), ""


def create_entity_timeindex_query(
    qdict,
    feature_store,
    column_alias_map=None,
):
    all_queries = []
    max_queries = []
    model_alias = "t"
    table_alias_count = 1
    _, table_name = feature_store.get_data_table_name(qdict["name"])
    full_name = qdict.get("table_name", None)
    if full_name is None:
        full_name = feature_store.db_schema + "." + table_name

    index, qualified_index = get_index(qdict)
    time_index, qualified_time_index = get_time_index_field(qdict)

    if time_index is None:
        time_index = "current_timestamp"
        qualified_time_index = "continual__time_index"

    for f in qdict["fields"]:
        if "relation" in f:
            relation = f["relation"]
            rel_time_index, _ = get_time_index_field(relation)
            if rel_time_index is not None:
                table_alias_count += 1
                index_query, max_query = render_index_query(
                    relation,
                    f["qualified_name"],
                    qualified_time_index,
                    feature_store,
                    column_alias_map=column_alias_map,
                    model_name=full_name,
                    table_alias=f"t{table_alias_count}",
                    model_alias=model_alias,
                )
                all_queries.append(index_query)
                if max_query != "":
                    max_queries.append(max_query)

            if "related" in relation:
                for related in relation["related"]:
                    rel_time_index, _ = get_time_index_field(related)
                    if rel_time_index is not None:
                        table_alias_count += 1
                        index_query, max_query = render_index_query(
                            related,
                            f["qualified_name"],
                            qualified_time_index,
                            feature_store,
                            column_alias_map=column_alias_map,
                            model_name=full_name,
                            table_alias=f"t{table_alias_count}",
                            model_alias=model_alias,
                        )
                        all_queries.append(index_query)
                        if max_query != "":
                            max_queries.append(max_query)

    if "related" in qdict:
        for related in qdict["related"]:
            rel_time_index, _ = get_time_index_field(related)
            if rel_time_index is not None:
                table_alias_count += 1
                index_query, max_query = render_index_query(
                    related,
                    qualified_index,
                    qualified_time_index,
                    feature_store,
                    column_alias_map=column_alias_map,
                    model_name=full_name,
                    table_alias=f"t{table_alias_count}",
                    model_alias=model_alias,
                )
                all_queries.append(index_query)
                if max_query != "":
                    max_queries.append(max_query)

    subquery_text = None
    maxquery_text = None
    if len(all_queries) > 0:
        if feature_store.get_type() == "databricks":
            subquery_text = " ".join(all_queries)
        else:
            subquery_text = ",".join(all_queries)

    if len(max_queries) > 0:
        maxquery_text = ",".join(max_queries)

    time_index_query = "WITH time_index AS "

    if feature_store.get_type() == "databricks":
        logging.info("subquery text: {}".format(subquery_text))
        logging.info("maxquery text: {}".format(maxquery_text))
        if subquery_text is not None:
            time_index_query += f"(SELECT {model_alias}.{index} as {qualified_index}, {model_alias}.{time_index} as {qualified_time_index}, {maxquery_text} FROM {full_name} as {model_alias} {subquery_text} group by {model_alias}.{index}, {model_alias}.{time_index})"
        else:
            time_index_query += f"(SELECT {index} as {qualified_index}, {time_index} as {qualified_time_index} FROM {full_name} as {model_alias})"
    else:
        if subquery_text is not None:
            time_index_query += f"(SELECT {index} as {qualified_index}, {time_index} as {qualified_time_index}, {subquery_text} FROM {full_name} as {model_alias})"
        else:
            time_index_query += f"(SELECT {index} as {qualified_index}, {time_index} as {qualified_time_index} FROM {full_name} as {model_alias})"

    return time_index_query


def create_select_list(
    qdict,
    match_index,
    alias_index,
    join_conditions,
    select_list,
    column_alias_map,
    feature_store,
    spine=False,
):
    project_id, table_name = feature_store.get_data_table_name(qdict["name"])
    index, qualified_index = get_index(qdict)
    time_index, qualified_time_index = get_time_index_field(qdict)

    if qualified_time_index is not None:
        qualified_time_index = qualified_time_index.replace(":", "_")
        qualified_time_index = qualified_time_index.replace("$", "_")

    fullname = qdict.get("table_name", None)
    if fullname is None:
        fullname = f"{feature_store.db_schema}.{table_name}"
    table_alias = f"t{alias_index}"

    join_condition = f"{match_index} = {table_alias}.{index}"
    if time_index is not None:

        join_condition += (
            f""" AND time_index.{qualified_time_index} = {table_alias}.{time_index} """
        )

    for f in qdict["fields"]:
        if spine and f["name"] == index:
            continue

        if f["name"] == time_index:
            continue
        col_name = f"t{alias_index}.{f['name']}"
        q_name = f["qualified_name"]
        alias_name = q_name
        if ":" in alias_name:
            alias_name = alias_name.replace(":", "_")
            alias_name = alias_name.replace("$", "_")

        select_list[col_name] = alias_name
        column_alias_map[alias_name] = q_name

    join_conditions[fullname] = (table_alias, join_condition)
    return


def create_training_query(
    query,
    feature_store,
    all_instances=False,
):
    logging.info("In Create training query")
    if isinstance(query, Query):
        qdict = query.to_dict()
    elif isinstance(query, management_types_pb2.Query):
        qdict = MessageToDict(query, preserving_proto_field_name=True)
    else:
        qdict = query

    # Columns are aliased because continual full qualified names are line fs1:feature1$fs2:feature2
    # This is not very database friendly. Hence we replace $ and : with _ and keep a mapping here.
    # This mapping is used to replaced the names in the return data.
    column_alias_map = {}

    join_conditions = {}
    select_list = {}
    filters = []

    time_index_query = create_entity_timeindex_query(
        qdict, feature_store=feature_store, column_alias_map=column_alias_map
    )
    index, _ = get_index(qdict)

    logging.info(f"time_index query: {time_index_query}")
    spine_alias_index = 1
    alias_index = 1

    create_select_list(
        qdict,
        f"time_index.{index}",
        alias_index,
        join_conditions,
        select_list,
        column_alias_map,
        feature_store,
        spine=True,
    )

    if "related" in qdict:
        for related in qdict["related"]:
            alias_index += 1
            create_select_list(
                related,
                f"time_index.{index}",
                alias_index,
                join_conditions,
                select_list,
                column_alias_map,
                feature_store,
                spine=True,
            )

    for f in qdict["fields"]:
        if "model" in f:
            if not all_instances:
                filters.append("t1." + f["name"] + " is NOT NULL")
            continue

        if "relation" not in f:
            continue

        relation = f["relation"]

        match_index = f"t{spine_alias_index}.{f['name']}"

        alias_index += 1
        create_select_list(
            relation,
            match_index,
            alias_index,
            join_conditions,
            select_list,
            column_alias_map,
            feature_store,
        )

        if "related" in relation:
            for related in relation["related"]:
                alias_index += 1
                create_select_list(
                    related,
                    match_index,
                    alias_index,
                    join_conditions,
                    select_list,
                    column_alias_map,
                    feature_store,
                    spine=True,
                )

    params = {
        "index_query": time_index_query,
        "select_list": select_list,
        "join_list": join_conditions,
        "filter_list": filters,
    }

    final_query = ""
    if time_index_query is None:
        final_query = non_time_train_template.render(params=params)
    else:
        final_query = train_template.render(params=params)
    logging.info(final_query)
    return final_query, column_alias_map
